from PyQt5.QtCore import QPointF

import torch
from ultralytics.utils.nms import TorchNMS

from coralnet_toolbox.Annotations.QtPolygonAnnotation import PolygonAnnotation
from coralnet_toolbox.Annotations.QtRectangleAnnotation import RectangleAnnotation

from coralnet_toolbox.utilities import simplify_polygon

from coralnet_toolbox.QtProgressBar import ProgressBar


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class ResultsProcessor:
    def __init__(self,
                 main_window,
                 class_mapping={},
                 uncertainty_thresh=None,
                 iou_thresh=None,
                 min_area_thresh=None,
                 max_area_thresh=None):
        self.main_window = main_window
        self.label_window = main_window.label_window
        self.image_window = main_window.image_window
        self.annotation_window = main_window.annotation_window

        self.class_mapping = class_mapping

        if uncertainty_thresh is None:
            uncertainty_thresh = self.main_window.get_uncertainty_thresh()

        if iou_thresh is None:
            iou_thresh = main_window.get_iou_thresh()

        if min_area_thresh is None:
            min_area_thresh = main_window.get_area_thresh_min()

        if max_area_thresh is None:
            max_area_thresh = main_window.get_area_thresh_max()

        self.uncertainty_thresh = uncertainty_thresh
        self.iou_thresh = iou_thresh
        self.min_area_thresh = min_area_thresh
        self.max_area_thresh = max_area_thresh
    
    def filter_by_uncertainty(self, results):
        """
        Filter the results based on the uncertainty threshold.
        """
        try:
            results = results[results.boxes.conf > self.uncertainty_thresh]
        except Exception as e:
            print(f"Warning: Failed to filter results by uncertainty\n{e}")

        return results

    def filter_by_iou(self, results):
        """Filter the results based on the IoU threshold."""
        try:
            results = results[TorchNMS.fast_nms(results.boxes.xyxy, results.boxes.conf, self.iou_thresh)]
        except Exception as e:
            print(f"Warning: Failed to filter results by IoU\n{e}")

        return results

    def filter_by_area(self, results):
        """
        Filter the results based on the area threshold.
        """
        try:
            x_norm, y_norm, w_norm, h_norm = results.boxes.xywhn.T
            area_norm = w_norm * h_norm
            results = results[(area_norm >= self.min_area_thresh) & (area_norm <= self.max_area_thresh)]
        except Exception as e:
            print(f"Warning: Failed to filter results by area\n{e}")

        return results

    def apply_filters_to_results(self, results):
        """Check if the results passed all filters."""
        results = self.filter_by_uncertainty(results)
        results = self.filter_by_iou(results)
        results = self.filter_by_area(results)
        return results

    def indices_pass_uncertainty(self, results):
        """
        Get the indices of results that pass the uncertainty threshold.
        """
        try:
            mask = results.boxes.conf > self.uncertainty_thresh
            indices = mask.nonzero().flatten().tolist()
        except Exception as e:
            print(f"Warning: Failed to get indices for uncertainty\n{e}")
            indices = []

        return indices

    def indices_pass_iou(self, results):
        """
        Get the indices of results that pass the IoU threshold.
        """
        try:
            indices = TorchNMS.fast_nms(results.boxes.xyxy, results.boxes.conf, self.iou_thresh).tolist()
        except Exception as e:
            print(f"Warning: Failed to get indices for IoU\n{e}")
            indices = []

        return indices

    def indices_pass_area(self, results):
        """
        Get the indices of results that pass the area threshold.
        """
        try:
            x_norm, y_norm, w_norm, h_norm = results.boxes.xywhn.T
            area_norm = w_norm * h_norm
            mask = (area_norm >= self.min_area_thresh) & (area_norm <= self.max_area_thresh)
            indices = mask.nonzero().flatten().tolist()
        except Exception as e:
            print(f"Warning: Failed to get indices for area\n{e}")
            indices = []

        return indices

    def indices_pass_filters(self, results):
        """
        Get the indices of results that pass all filters.
        """
        indices_uncertainty = set(self.indices_pass_uncertainty(results))
        indices_iou = set(self.indices_pass_iou(results))
        indices_area = set(self.indices_pass_area(results))

        # Get the indexes of results that pass all filters (intersection)
        indices = indices_uncertainty.intersection(indices_iou).intersection(indices_area)

        return list(indices)
    
    def _process_single_result_set(self, results, model_type):
        """
        Processes a *single* Results object from a single work area.
        
        This function filters the results, runs NMS against existing annotations,
        and creates a list of new annotation objects to be added.
        """
        annotations_to_add = []
        new_detections = []
        
        if not results or not results.boxes: 
            return annotations_to_add
            
        image_path = results.path.replace("\\", "/")
        
        # --- 1. Filter new results by uncertainty and area ---
        indices_uncertainty = set(self.indices_pass_uncertainty(results))
        indices_area = set(self.indices_pass_area(results))
        filtered_indices = list(indices_uncertainty.intersection(indices_area))
        
        for idx in filtered_indices:
            try:
                box = results.boxes[idx]
                xyxy = box.xyxy[0].cpu().numpy()
                conf = float(box.conf[0].cpu().numpy())
                cls_id = int(box.cls[0].cpu().numpy())
                cls_name = results.names[cls_id]
                
                bbox_coords = [float(xyxy[0]), float(xyxy[1]), float(xyxy[2]), float(xyxy[3])]

                detection_data = {
                    'bbox': bbox_coords,
                    'confidence': conf,
                    'class_name': cls_name,
                    'class_id': cls_id,
                    'is_existing': False,
                    'polygon_points': None,
                    'model_type': model_type
                }
                
                if model_type == 'segmentation':
                    if results.masks and idx < len(results.masks.xy):
                        xy = results.masks.xy[idx]
                        points = [(float(x), float(y)) for x, y in xy]
                        detection_data['polygon_points'] = points
                        new_detections.append(detection_data)
                else:  # 'detection'
                    new_detections.append(detection_data)
                    
            except Exception as e:
                print(f"Warning: Failed to extract result in first pass: {e}")
        
        if not new_detections:
            return annotations_to_add
            
        # --- 2. Run NMS vs. Existing Annotations (Class-Agnostic) ---
        
        existing_annotations = self.annotation_window.get_image_annotations(image_path)
        existing_annotations = [a for a in existing_annotations if (isinstance(a, RectangleAnnotation) or 
                                                                    isinstance(a, PolygonAnnotation))]
        existing_detections = self._convert_annotations_to_nms_format(existing_annotations)
        
        # Combine existing (with 1.0 conf) and new survivors
        all_detections_combined = existing_detections + new_detections
        
        if not all_detections_combined:
            return annotations_to_add
            
        # Convert to tensors for NMS
        bboxes = torch.tensor([det['bbox'] for det in all_detections_combined], dtype=torch.float32)
        scores = torch.tensor([det['confidence'] for det in all_detections_combined], dtype=torch.float32)
        
        try:
            # Apply NMS using TorchNMS. This is class-agnostic.
            keep_indices_tensor = TorchNMS.fast_nms(bboxes, scores, self.iou_thresh)
            keep_indices = keep_indices_tensor.tolist()  # Convert tensor to standard list of ints
        except Exception as e:
            print(f"Warning: Stage 2 NMS (TorchNMS) failed: {e}")
            keep_indices = list(range(len(all_detections_combined)))
            
        # --- 3. Create Annotations for final survivors ---
        
        for idx in keep_indices:
            # Check if the kept index is a NEW detection
            original_detection = all_detections_combined[idx]
            if not original_detection['is_existing']:
                try:
                    conf = original_detection['confidence']
                    cls_name = original_detection['class_name']
                    
                    short_label = self.get_mapped_short_label(cls_name, conf)
                    label = self.label_window.get_label_by_short_code(short_label)
                    
                    annotation = None
                    if original_detection['model_type'] == 'detection':
                        xmin, ymin, xmax, ymax = original_detection['bbox']
                        annotation = self.create_rectangle_annotation(xmin, ymin, xmax, ymax, label, image_path)
                    
                    elif original_detection['model_type'] == 'segmentation' and original_detection['polygon_points']:
                        points = simplify_polygon(original_detection['polygon_points'], 0.1) 
                        annotation = self.create_polygon_annotation(points, label, image_path)

                    if annotation:
                        processed_annotation = self._post_process_new_annotation(annotation, cls_name, conf)
                        annotations_to_add.append(processed_annotation)
                
                except Exception as e:
                    print(f"Warning: Failed to create annotation: {e}")
        
        return annotations_to_add

    def _convert_annotations_to_nms_format(self, annotations):
        """Convert existing annotations to NMS format using their built-in method."""
        nms_detections = []
        
        for annotation in annotations:
            try:
                # Use the annotation's built-in NMS conversion method
                nms_detection = annotation.to_nms_detection()
                # Override confidence to 1.0 to ensure existing annotations
                # always win NMS against new predictions (which are <= 1.0).
                nms_detection['confidence'] = 1.0
                nms_detections.append(nms_detection)
            except Exception as e:
                print(f"Warning: Failed to convert annotation {annotation.id} to NMS format: {e}")
        
        return nms_detections
        
    def get_mapped_short_label(self, cls_name, conf):
        """
        Get the short label for a detection result based on confidence and class mapping.

        :param cls_name: Class name
        :param conf: Confidence score
        :return: Short label as a string
        """
        if conf <= self.uncertainty_thresh:
            return 'Review'
        return self.class_mapping.get(cls_name, {}).get('short_label_code', 'Review')

    def create_rectangle_annotation(self, xmin, ymin, xmax, ymax, label, image_path):
        """
        Create a rectangle annotation for the given bounding box coordinates and label.

        :param xmin: Minimum x-coordinate
        :param ymin: Minimum y-coordinate
        :param xmax: Maximum x-coordinate
        :param ymax: Maximum y-coordinate
        :param label: Label object
        :param image_path: Path to the image
        :return: RectangleAnnotation object
        """
        try:
            top_left = QPointF(xmin, ymin)
            bottom_right = QPointF(xmax, ymax)
            annotation = RectangleAnnotation(top_left,
                                             bottom_right,
                                             label.short_label_code,
                                             label.long_label_code,
                                             label.color,
                                             image_path,
                                             label.id,
                                             self.main_window.get_transparency_value())
        except Exception:
            annotation = None

        return annotation

    def create_polygon_annotation(self, points, label, image_path):
        """
        Create a polygon annotation for the given points and label.

        :param points: List of polygon points
        :param label: Label object
        :param image_path: Path to the image
        :return: PolygonAnnotation object
        """
        try:
            points = [QPointF(x, y) for x, y in points]
            annotation = PolygonAnnotation(points,
                                           label.short_label_code,
                                           label.long_label_code,
                                           label.color,
                                           image_path,
                                           label.id,
                                           self.main_window.get_transparency_value())
        except Exception:
            annotation = None

        return annotation

    def extract_classification_result(self, result):
        """
        Extract relevant information from a classification result.
        """
        predictions = {}

        try:
            image_path = result.path.replace("\\", "/")
            class_names = result.names
            top1 = result.probs.top1
            top1conf = result.probs.top1conf
            top1cls = class_names[top1]
            top5 = result.probs.top5
            top5conf = result.probs.top5conf
        except Exception as e:
            print(f"Warning: Failed to process classification result\n{e}")
            return predictions

        for idx, conf in zip(top5, top5conf):
            class_name = class_names[idx]
            label = self.label_window.get_label_by_short_code(class_name)
            if label:
                predictions[label] = float(conf)

        return image_path, top1cls, top1conf, predictions

    def process_classification_results(self, results_generator, annotations, progress_bar=None):
        """
        Process the classification results from the results generator.
        
        Args:
            results_generator: Generator of classification results
            annotations: List of annotations to update
            progress_bar: Optional external progress bar. If None, creates its own.
        """
        # Track if we created the progress bar ourselves
        progress_bar_created_here = progress_bar is None
        
        if progress_bar is None:
            progress_bar = ProgressBar(self.annotation_window, title="Making Classification Predictions")
            progress_bar.show()
        
        progress_bar.start_progress(len(annotations))

        try:
            for result, annotation in zip(results_generator, annotations):
                if result:
                    try:
                        # Extract results and pass them to the dedicated update/display method
                        _, cls_name, conf, predictions = self.extract_classification_result(result)
                        self._update_and_display_classification(annotation, cls_name, conf, predictions)
                    except Exception as e:
                        print(f"Warning: Failed to process classification result for annotation {annotation.id}\n{e}")
                
                # Only update progress if we have a progress bar (external or internal)
                if progress_bar:
                    progress_bar.update_progress()

        finally:
            # Only close progress bar if we created it ourselves
            if progress_bar_created_here and progress_bar:
                progress_bar.stop_progress()
                progress_bar.close()
        
    def _update_and_display_classification(self, annotation, cls_name, conf, predictions):
        """
        Updates an existing annotation with classification results and refreshes the UI if visible.

        :param annotation: Annotation object to update.
        :param cls_name: The top predicted class name.
        :param conf: The top confidence score.
        :param predictions: Dictionary of all class predictions.
        """
        # Update the machine confidence values with all predictions
        annotation.update_machine_confidence(predictions)

        # Determine the final label, setting it to 'Review' if confidence is too low
        final_label = self.label_window.get_label_by_short_code(cls_name)
        if conf < self.uncertainty_thresh:
            final_label = self.label_window.get_label_by_id('-1')
        
        # Update the annotation's label
        if final_label:
            annotation.update_label(final_label)

        # If the annotation's image is currently displayed, refresh its visual state
        if annotation.image_path == self.annotation_window.current_image_path:
            annotation.create_cropped_image(self.annotation_window.rasterio_image)
            # Update the confidence window if this annotation is selected
            if annotation in self.annotation_window.selected_annotations:
                self.main_window.confidence_window.display_cropped_image(annotation)
        
        # Update the annotation count in the image table
        self.image_window.update_image_annotations(annotation.image_path)

    def extract_detection_result(self, result):
        """
        Extract relevant information from a detection result.

        :param result: Detection result
        :return: Tuple containing class, class name, confidence, and bounding box coordinates
        """
        # Class ID, class name, confidence, and bounding box coordinates
        image_path = result.path.replace("\\", "/")
        cls = int(result.boxes.cls.cpu().numpy()[0])
        cls_name = result.names[cls]
        conf = float(result.boxes.conf.cpu().numpy()[0])
        xmin, ymin, xmax, ymax = map(float, result.boxes.xyxy.cpu().numpy()[0])

        return image_path, cls, cls_name, conf, xmin, ymin, xmax, ymax
    
    def extract_segmentation_result(self, result):
        """
        Extract relevant information from a segmentation result.

        :param result: Segmentation result
        :return: Tuple containing class, class name, confidence, and polygon points
        """
        # Class ID, class name, confidence, and polygon points
        image_path = result.path.replace("\\", "/")
        cls = int(result.boxes.cls.cpu().numpy()[0])
        cls_name = result.names[cls]
        conf = float(result.boxes.conf.cpu().numpy()[0])

        return image_path, cls, cls_name, conf

    def process_segmentation_results(self, results_list):
        """
        Process segmentation results. Loops through the list and processes
        one Results object at a time.
        """
        all_new_annos = []
        image_path = None
        
        for results in results_list:
            if not results or not results.boxes:
                continue
            if image_path is None:
                image_path = results.path.replace("\\", "/")
                
            new_annos = self._process_single_result_set(results, 'segmentation')
            all_new_annos.extend(new_annos)
            
        if all_new_annos:
            self.annotation_window.add_annotations(all_new_annos)
            if self.annotation_window.current_image_path == image_path:
                self.annotation_window.load_annotations()

    def process_detection_results(self, results_list):
        """
        Process detection results. Loops through the list and processes
        one Results object at a time.
        """
        all_new_annos = []
        image_path = None
        
        for results in results_list:
            if not results or not results.boxes:
                continue
            if image_path is None:
                image_path = results.path.replace("\\", "/")
                
            new_annos = self._process_single_result_set(results, 'detection')
            all_new_annos.extend(new_annos)
            
        if all_new_annos:
            self.annotation_window.add_annotations(all_new_annos)
            if self.annotation_window.current_image_path == image_path:
                self.annotation_window.load_annotations()
        
    def _post_process_new_annotation(self, annotation, cls_name, conf):
        """
        Applies final processing to a newly created annotation before batch addition.
        
        :param annotation: The annotation object to process.
        :param cls_name: The original class name from the model.
        :param conf: The confidence score from the model.
        :return: The processed annotation object.
        """
        # Update the confidence values for the prediction
        predictions = {self.label_window.get_label_by_short_code(cls_name): conf}
        annotation.update_machine_confidence(predictions)

        # If the confidence is below the threshold, update the label to 'Review'
        if conf < self.uncertainty_thresh:
            review_label = self.label_window.get_label_by_id('-1')
            annotation.update_label(review_label)
            annotation.set_verified(False)
        
        return annotation


import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (QFileDialog, QApplication, QMessageBox, QVBoxLayout,
                             QLabel, QLineEdit, QDialog, QHBoxLayout, QPushButton,
                             QComboBox, QFormLayout, QGroupBox, QScrollArea, QWidget)

from ultralytics import YOLO

from coralnet_toolbox.Icons import get_icon


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class Base(QDialog):
    """
    A dialog for optimizing and exporting a YOLO model with custom parameters.
    """

    def __init__(self, main_window, parent=None):
        """
        Initialize the dialog with the main window and optional parent widget.

        param main_window: The main window of the application.
        param parent: The parent widget of the dialog.
        """
        super().__init__(parent)
        self.main_window = main_window
        
        self.setWindowIcon(get_icon("coral.png"))
        self.setWindowTitle("Optimize Model")
        self.resize(400, 325)

        self.model_path = ""
        self.custom_params = []

        self.layout = QVBoxLayout(self)
        
        # Setup the information layout
        self.setup_info_layout()
        # Setup the model layout
        self.setup_model_layout()
        # Setup the parameters layout
        self.setup_parameters_layout()
        # Setup the buttons layout
        self.setup_buttons_layout()
        
    def setup_info_layout(self):
        """Setup information layout."""
        group_box = QGroupBox("Information")
        layout = QVBoxLayout()

        # Create a QLabel with explanatory text and hyperlink
        info_label = QLabel("Details on different production formats can be found "
                            "<a href='https://docs.ultralytics.com/modes/export/#export-formats'>here</a>.")
        
        info_label.setOpenExternalLinks(True)
        info_label.setWordWrap(True)
        layout.addWidget(info_label)
        
        group_box.setLayout(layout)
        self.layout.addWidget(group_box)
        
    def setup_model_layout(self):
        """Setup model layout."""
        group_box = QGroupBox("Model")
        layout = QFormLayout()

        # Model file selection layout
        model_layout = QHBoxLayout()
        self.model_path_edit = QLineEdit()
        self.model_browse_button = QPushButton("Browse...")
        self.model_browse_button.clicked.connect(self.browse_file)
        model_layout.addWidget(self.model_path_edit)
        model_layout.addWidget(self.model_browse_button)
        layout.addRow("Model Path:", model_layout)
        
        # Export Format Dropdown
        self.export_format_combo = QComboBox()
        self.export_format_combo.addItems(["torchscript", "onnx", "openvino", "engine"])
        self.export_format_combo.setEditable(True)
        layout.addRow("Export Format:", self.export_format_combo)

        # Set the layout to the group box and add to main layout
        group_box.setLayout(layout)
        self.layout.addWidget(group_box)
    
    def setup_parameters_layout(self):
        """Setup parameters layout."""
        # Create parameters group box
        group_box = QGroupBox("Parameters")
        group_layout = QVBoxLayout(group_box)

        # Create a widget to hold the form layout
        form_widget = QWidget()
        form_layout = QVBoxLayout(form_widget)

        # Custom parameters section (inside form layout)
        self.custom_params_layout = QVBoxLayout()
        form_layout.addLayout(self.custom_params_layout)

        # Create the scroll area
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setWidget(form_widget)
        group_layout.addWidget(scroll_area)

        # Buttons for parameter management (outside scroll area)
        buttons_layout = QHBoxLayout()
        self.add_param_button = QPushButton("Add Parameter")
        self.add_param_button.clicked.connect(self.add_parameter_pair)
        remove_button = QPushButton("Remove Parameter")
        remove_button.clicked.connect(self.remove_last_parameter)
        buttons_layout.addWidget(self.add_param_button)
        buttons_layout.addWidget(remove_button)
        group_layout.addLayout(buttons_layout)

        self.layout.addWidget(group_box)

    def setup_buttons_layout(self):
        """Setup buttons layout."""
        cancel_button = QPushButton("Cancel")
        cancel_button.clicked.connect(self.reject)
        self.layout.addWidget(cancel_button)
        
        accept_button = QPushButton("Accept")
        accept_button.clicked.connect(self.optimize_model)
        self.layout.addWidget(accept_button)

    def add_parameter_pair(self):
        """
        Add a new pair of parameter name and value input fields.
        """
        param_layout = QHBoxLayout()
        param_name = QLineEdit()
        param_value = QLineEdit()
        param_layout.addWidget(param_name)
        param_layout.addWidget(param_value)

        self.custom_params.append((param_name, param_value))
        self.custom_params_layout.addLayout(param_layout)
        
    def remove_last_parameter(self):
        """
        Remove the last parameter pair from the layout.
        """
        if self.custom_params:
            # Get the last parameter pair
            param_name, param_value = self.custom_params.pop()
            # Get the layout containing these widgets
            layout_to_remove = self.custom_params_layout.takeAt(self.custom_params_layout.count() - 1)
            # Delete the widgets
            while layout_to_remove.count():
                widget = layout_to_remove.takeAt(0).widget()
                widget.deleteLater()
            # Delete the layout
            layout_to_remove.deleteLater()

    def browse_file(self):
        """
        Open a file dialog to select a model file and display its path.
        """
        options = QFileDialog.Options()
        file_path, _ = QFileDialog.getOpenFileName(self, 
                                                   "Open Model File", 
                                                   "", 
                                                   "Model Files (*.pt)", 
                                                   options=options)
        if file_path:
            if ".pt" in file_path or ".pth" in file_path:
                self.model_path = file_path
                self.model_path_edit.setText(file_path)
            else:
                QMessageBox.critical(self, "Error", "Please select a .pt model file.")
                self.model_path_edit.setText("")

    def accept(self):
        """
        Override the accept method to optimize the model before closing the dialog.
        """
        self.optimize_model()
        super().accept()

    def get_optimization_parameters(self):
        """
        Extract and return the optimization parameters from the dialog widgets.
        """
        params = {'format': self.export_format_combo.currentText()}

        for param_name, param_value in self.custom_params:
            name = param_name.text().strip()
            value = param_value.text().strip().lower()
            if name:
                if value == 'true':
                    params[name] = True
                elif value == 'false':
                    params[name] = False
                else:
                    try:
                        params[name] = int(value)
                    except ValueError:
                        try:
                            params[name] = float(value)
                        except ValueError:
                            params[name] = value

        return params

    def optimize_model(self):
        """
        Optimize and export the model using the specified parameters.
        """
        params = self.get_optimization_parameters()

        # Set the cursor to waiting (busy) cursor
        QApplication.setOverrideCursor(Qt.WaitCursor)

        try:
            # Initialize the model and export with given params
            YOLO(self.model_path).export(**params)

            message = "Model export successful."
            QMessageBox.information(self, "Model Export Status", message)

        except Exception as e:
            # Display an error message box to the user
            error_message = f"An error occurred when converting model: {e}"
            QMessageBox.critical(self, "Error", error_message)
            print(error_message)

        # Restore the cursor to the default cursor
        QApplication.restoreOverrideCursor()


import psutil
import GPUtil
import collections
import time
import platform

import pyqtgraph as pg
from PyQt5.QtWidgets import QMainWindow, QGridLayout, QWidget, QLabel
from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtGui import QFont

from coralnet_toolbox.Icons import get_icon


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class SystemMonitor(QMainWindow):
    """
    A PyQt5 application to monitor and display real-time CPU, Memory, and GPU usage.
    The monitoring starts/stops automatically when the window is shown/hidden.
    """
    def __init__(self):
        super().__init__()

        # --- Window Properties ---
        self.setWindowTitle("System Monitor")
        self.setWindowIcon(get_icon("system_monitor.png"))
        self.setGeometry(100, 100, 800, 900)

        # --- Data Storage ---
        # Use collections.deque for efficient, fixed-size data storage
        self.history_size = 100  # Number of data points to display
        self.cpu_data = collections.deque([0] * self.history_size, maxlen=self.history_size)
        self.mem_data = collections.deque([0] * self.history_size, maxlen=self.history_size)
        self.gpu_data = collections.deque([0] * self.history_size, maxlen=self.history_size)
        
        # Per-core CPU data
        self.cpu_cores = psutil.cpu_count(logical=True)
        self.per_core_data = []
        for _ in range(self.cpu_cores):
            self.per_core_data.append(collections.deque([0] * self.history_size, maxlen=self.history_size))

        # Disk I/O data
        self.disk_read_data = collections.deque([0] * self.history_size, maxlen=self.history_size)
        self.disk_write_data = collections.deque([0] * self.history_size, maxlen=self.history_size)
        self.prev_disk_io = psutil.disk_io_counters()
        self.prev_time = time.monotonic()
        
        # Network traffic data
        self.net_sent_data = collections.deque([0] * self.history_size, maxlen=self.history_size)
        self.net_recv_data = collections.deque([0] * self.history_size, maxlen=self.history_size)
        self.prev_net_io = psutil.net_io_counters()
        
        # X-axis data (time steps)
        self.x_axis = list(range(self.history_size))

        # --- Initialize UI ---
        self.setup_layout()

        # --- Timer for Real-Time Updates ---
        # Set up a QTimer to trigger the update_plots method every 1000 ms (1 second)
        # We do NOT start the timer here. It will be started by the showEvent.
        self.timer = QTimer()
        self.timer.setInterval(1000)
        self.timer.timeout.connect(self.update_plots)

    def setup_layout(self):
        """
        Initializes the user interface, setting up the layout and plots.
        """
        # --- Central Widget and Layout ---
        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        layout = QGridLayout(central_widget)
        layout.setVerticalSpacing(20)  # Add more vertical spacing between plots

        # --- Styling ---
        pg.setConfigOption('background', 'w')  # White background for graphs
        pg.setConfigOption('foreground', '#000000')  # Black foreground elements in graphs
        plot_label_style = {"color": "#000000", "font-size": "14pt"}  # Black text for plot labels
        title_style = {"color": "#000000", "font-size": "16pt", "font-weight": "bold"}  # Black text

        # --- Stats Labels - Only Hardware Info ---
        self.stats_widget = QWidget()
        stats_layout = QGridLayout(self.stats_widget)
        stats_layout.setHorizontalSpacing(20)  # Add horizontal spacing between items
        stats_layout.setContentsMargins(10, 10, 10, 20)  # Add margins around the widget

        # Get system hardware information
        cpu_info = platform.processor() or "CPU"
        total_memory = round(psutil.virtual_memory().total / (1024**3), 1)
        
        try:
            gpus = GPUtil.getGPUs()
            gpu_info = gpus[0].name if gpus else "No GPU detected"
        except Exception:
            gpu_info = "GPU information unavailable"
        
        # Hardware info labels with consistent styling and alignment
        info_label_style = "color: black; font-size: 11pt; font-weight: bold; padding: 5px;"

        self.cpu_info_label = QLabel(f"CPU: {cpu_info}")
        self.cpu_info_label.setStyleSheet(info_label_style)
        self.cpu_info_label.setAlignment(Qt.AlignCenter)
        stats_layout.addWidget(self.cpu_info_label, 0, 0)
        
        self.mem_info_label = QLabel(f"Memory: {total_memory:.1f} GB")
        self.mem_info_label.setStyleSheet(info_label_style)
        self.mem_info_label.setAlignment(Qt.AlignCenter)
        stats_layout.addWidget(self.mem_info_label, 0, 1)
        
        self.gpu_info_label = QLabel(f"GPU: {gpu_info}")
        self.gpu_info_label.setStyleSheet(info_label_style)
        self.gpu_info_label.setAlignment(Qt.AlignCenter)
        stats_layout.addWidget(self.gpu_info_label, 0, 2)
        
        # Set column stretching to distribute space evenly
        stats_layout.setColumnStretch(0, 1)
        stats_layout.setColumnStretch(1, 1)
        stats_layout.setColumnStretch(2, 1)

        layout.addWidget(self.stats_widget, 0, 0, 1, 2)

        # Helper function to add bottom space for x-axis and enable auto-range
        def configure_plot(plot_widget, fixed_y_range=None):
            plot_widget.getPlotItem().getAxis('bottom').setHeight(40)
            plot_widget.getPlotItem().layout.setContentsMargins(10, 10, 10, 30)
            
            # If fixed_y_range is None, enable auto-ranging, otherwise set fixed range
            if fixed_y_range is None:
                plot_widget.enableAutoRange(axis='y')
            else:
                min_val, max_val = fixed_y_range
                plot_widget.setYRange(min_val, max_val, padding=0.05)
                
            return plot_widget
            
        # Helper function to create a value label for a plot
        def create_value_label(plot_widget, initial_text="Current: N/A"):
            text_item = pg.TextItem(text=initial_text, color='#000000', anchor=(0, 0))
            font = QFont()
            font.setBold(True)
            text_item.setFont(font)
            plot_widget.addItem(text_item)
            # Position at top-left corner with some margin
            text_item.setPos(10, 10)
            return text_item

        # --- CPU Plot --- (fixed 0-100 range)
        self.cpu_plot_widget = pg.PlotWidget()
        self.cpu_plot_widget.setTitle("CPU Usage (%)", **title_style)
        self.cpu_plot_widget.setLabel("left", "Usage", units="%", **plot_label_style)
        self.cpu_plot_widget.setLabel("bottom", "Time (s)", **plot_label_style)
        self.cpu_plot_widget.showGrid(x=True, y=True, alpha=0.3)
        self.cpu_curve = self.cpu_plot_widget.plot(self.x_axis, list(self.cpu_data), pen=pg.mkPen('#00A3FF', width=3))
        self.cpu_value_label = create_value_label(self.cpu_plot_widget, "Current: 0%")
        configure_plot(self.cpu_plot_widget, fixed_y_range=(0, 100))
        layout.addWidget(self.cpu_plot_widget, 1, 0)

        # --- Per-Core CPU Plot --- (fixed 0-100 range)
        self.core_plot_widget = pg.PlotWidget()
        self.core_plot_widget.setTitle("Per-Core CPU Usage (%)", **title_style)
        self.core_plot_widget.setLabel("left", "Usage", units="%", **plot_label_style)
        self.core_plot_widget.setLabel("bottom", "Time (s)", **plot_label_style)
        self.core_plot_widget.showGrid(x=True, y=True, alpha=0.3)
        
        self.per_core_curves = []
        for i in range(self.cpu_cores):
            color = pg.intColor(i, hues=self.cpu_cores)
            self.per_core_curves.append(
                self.core_plot_widget.plot(
                    self.x_axis, 
                    list(self.per_core_data[i]), 
                    pen=pg.mkPen(color, width=2)
                )
            )
        configure_plot(self.core_plot_widget, fixed_y_range=(0, 100))
        layout.addWidget(self.core_plot_widget, 1, 1)

        # --- Memory Plot --- (fixed 0-100 range)
        self.mem_plot_widget = pg.PlotWidget()
        self.mem_plot_widget.setTitle("Memory Usage (%)", **title_style)
        self.mem_plot_widget.setLabel("left", "Usage", units="%", **plot_label_style)
        self.mem_plot_widget.setLabel("bottom", "Time (s)", **plot_label_style)
        self.mem_plot_widget.showGrid(x=True, y=True, alpha=0.3)
        self.mem_curve = self.mem_plot_widget.plot(self.x_axis, list(self.mem_data), pen=pg.mkPen('#FF5733', width=3))
        self.mem_value_label = create_value_label(self.mem_plot_widget, "Current: 0%")
        configure_plot(self.mem_plot_widget, fixed_y_range=(0, 100))
        layout.addWidget(self.mem_plot_widget, 2, 0)

        # --- Disk I/O Plot --- (auto-ranging)
        self.disk_plot_widget = pg.PlotWidget()
        self.disk_plot_widget.setTitle("Disk I/O (MB/s)", **title_style)
        self.disk_plot_widget.setLabel("left", "Transfer Rate", units="MB/s", **plot_label_style)
        self.disk_plot_widget.setLabel("bottom", "Time (s)", **plot_label_style)
        self.disk_plot_widget.showGrid(x=True, y=True, alpha=0.3)
        
        self.disk_read_curve = self.disk_plot_widget.plot(
            self.x_axis, 
            list(self.disk_read_data), 
            pen=pg.mkPen('#FFDD33', width=3)
        )
        self.disk_write_curve = self.disk_plot_widget.plot(
            self.x_axis, 
            list(self.disk_write_data), 
            pen=pg.mkPen('#33DDFF', width=3)
        )
        
        self.disk_value_label = create_value_label(self.disk_plot_widget, "R: 0 MB/s, W: 0 MB/s")
        configure_plot(self.disk_plot_widget, fixed_y_range=None)  # Auto-range
        layout.addWidget(self.disk_plot_widget, 2, 1)

        # --- GPU Plot --- (fixed 0-100 range)
        self.gpu_plot_widget = pg.PlotWidget()
        self.gpu_plot_widget.setTitle("GPU Usage (%)", **title_style)
        self.gpu_plot_widget.setLabel("left", "Usage", units="%", **plot_label_style)
        self.gpu_plot_widget.setLabel("bottom", "Time (s)", **plot_label_style)
        self.gpu_plot_widget.showGrid(x=True, y=True, alpha=0.3)
        self.gpu_curve = self.gpu_plot_widget.plot(self.x_axis, list(self.gpu_data), pen=pg.mkPen('#33FF57', width=3))
        self.gpu_value_label = create_value_label(self.gpu_plot_widget, "Current: 0%")
        configure_plot(self.gpu_plot_widget, fixed_y_range=(0, 100))
        layout.addWidget(self.gpu_plot_widget, 3, 0)

        # --- Network Traffic Plot --- (auto-ranging)
        self.net_plot_widget = pg.PlotWidget()
        self.net_plot_widget.setTitle("Network Traffic (MB/s)", **title_style)
        self.net_plot_widget.setLabel("left", "Transfer Rate", units="MB/s", **plot_label_style)
        self.net_plot_widget.setLabel("bottom", "Time (s)", **plot_label_style)
        self.net_plot_widget.showGrid(x=True, y=True, alpha=0.3)
        
        self.net_sent_curve = self.net_plot_widget.plot(
            self.x_axis, 
            list(self.net_sent_data), 
            pen=pg.mkPen('#FF33DD', width=3)
        )
        self.net_recv_curve = self.net_plot_widget.plot(
            self.x_axis, 
            list(self.net_recv_data), 
            pen=pg.mkPen('#DD33FF', width=3)
        )
        
        self.net_value_label = create_value_label(self.net_plot_widget, "Up: 0 MB/s, Down: 0 MB/s")
        configure_plot(self.net_plot_widget, fixed_y_range=None)  # Auto-range
        layout.addWidget(self.net_plot_widget, 3, 1)

    def update_plots(self):
        """
        Fetches new system stats and updates the plot data.
        This method is called by the QTimer.
        """
        curr_time = time.monotonic()
        time_delta = curr_time - self.prev_time
        
        # --- Get CPU, Memory and GPU percentages ---
        # CPU usage
        cpu_percent = psutil.cpu_percent()
        self.cpu_data.append(cpu_percent)
        self.cpu_value_label.setText(f"Current: {cpu_percent:.1f}%")
        
        # Per-core CPU usage
        per_core = psutil.cpu_percent(percpu=True)
        for i, usage in enumerate(per_core):
            if i < len(self.per_core_data):
                self.per_core_data[i].append(usage)
                self.per_core_curves[i].setData(self.x_axis, list(self.per_core_data[i]))

        # Memory usage
        mem_percent = psutil.virtual_memory().percent
        self.mem_data.append(mem_percent)
        self.mem_value_label.setText(f"Current: {mem_percent:.1f}%")

        # GPU usage
        try:
            gpus = GPUtil.getGPUs()
            if gpus:
                gpu_percent = gpus[0].load * 100
                self.gpu_value_label.setText(f"Current: {gpu_percent:.1f}%")
            else:
                gpu_percent = 0
                self.gpu_value_label.setText("Current: 0%")
        except Exception:
            gpu_percent = 0
            self.gpu_value_label.setText("Current: 0%")

        self.gpu_data.append(gpu_percent)
        
        # Disk I/O
        curr_disk_io = psutil.disk_io_counters()
        
        # Calculate read/write rates in MB/s
        if curr_disk_io and self.prev_disk_io:
            read_bytes = curr_disk_io.read_bytes - self.prev_disk_io.read_bytes
            write_bytes = curr_disk_io.write_bytes - self.prev_disk_io.write_bytes
            read_mb_per_sec = (read_bytes / time_delta) / (1024**2)
            write_mb_per_sec = (write_bytes / time_delta) / (1024**2)
            
            self.disk_read_data.append(read_mb_per_sec)
            self.disk_write_data.append(write_mb_per_sec)
            
            # Update the value label with current values
            self.disk_value_label.setText(f"R: {read_mb_per_sec:.1f} MB/s, W: {write_mb_per_sec:.1f} MB/s")
            
            # Let pyqtgraph handle auto-ranging automatically
            self.disk_plot_widget.enableAutoRange(axis='y')
    
        self.prev_disk_io = curr_disk_io
        
        # Network traffic
        curr_net_io = psutil.net_io_counters()
        if curr_net_io and self.prev_net_io:
            sent_bytes = curr_net_io.bytes_sent - self.prev_net_io.bytes_sent
            recv_bytes = curr_net_io.bytes_recv - self.prev_net_io.bytes_recv
            
            sent_mb_per_sec = (sent_bytes / time_delta) / (1024**2)
            recv_mb_per_sec = (recv_bytes / time_delta) / (1024**2)
            
            self.net_sent_data.append(sent_mb_per_sec)
            self.net_recv_data.append(recv_mb_per_sec)
            
            # Update the value label with current values
            self.net_value_label.setText(f"Up: {sent_mb_per_sec:.1f} MB/s, Down: {recv_mb_per_sec:.1f} MB/s")
            
            # Let pyqtgraph handle auto-ranging automatically
            self.net_plot_widget.enableAutoRange(axis='y')
    
        self.prev_net_io = curr_net_io

        # --- Update Plot Curves ---
        # Convert deque to list for plotting
        self.cpu_curve.setData(self.x_axis, list(self.cpu_data))
        self.mem_curve.setData(self.x_axis, list(self.mem_data))
        self.gpu_curve.setData(self.x_axis, list(self.gpu_data))
        self.disk_read_curve.setData(self.x_axis, list(self.disk_read_data))
        self.disk_write_curve.setData(self.x_axis, list(self.disk_write_data))
        self.net_sent_curve.setData(self.x_axis, list(self.net_sent_data))
        self.net_recv_curve.setData(self.x_axis, list(self.net_recv_data))

        self.prev_time = curr_time

    # --- Event Handlers to Control Monitoring ---
    
    def showEvent(self, event):
        """
        Overrides the QWidget's showEvent.
        Starts the timer when the window is shown.
        """
        super().showEvent(event)
        if not self.timer.isActive():
            self.timer.start()

    def hideEvent(self, event):
        """
        Overrides the QWidget's hideEvent.
        Stops the timer when the window is hidden (e.g., minimized).
        """
        super().hideEvent(event)
        if self.timer.isActive():
            self.timer.stop()

    def closeEvent(self, event):
        """
        Overrides the QMainWindow's closeEvent.
        Ensures the timer is stopped when the window is closed.
        """
        if self.timer.isActive():
            self.timer.stop()
        super().closeEvent(event)


import warnings

from PyQt5.QtCore import Qt, QRectF
from PyQt5.QtGui import QMouseEvent, QPen, QColor, QBrush
from PyQt5.QtWidgets import (QGraphicsRectItem, QMessageBox, QGraphicsPixmapItem)

from coralnet_toolbox.Tools.QtTool import Tool
from coralnet_toolbox.QtWorkArea import WorkArea

warnings.filterwarnings("ignore", category=DeprecationWarning)


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class WorkAreaTool(Tool):
    """Tool for creating and managing work areas (rectangular regions) on an image."""
    
    def __init__(self, annotation_window):
        """Initialize the work area tool with the annotation window."""
        super().__init__(annotation_window)
        self.cursor = Qt.CrossCursor
        self.default_cursor = Qt.ArrowCursor  # Add this for clarity

        # State for drawing work areas
        self.drawing = False
        self.start_pos = None
        self.current_rect = None
        self.work_areas = []  # List to store WorkArea objects for the current image
        
        # Style settings for drawing the work area rectangle - update to use blue dashed line
        self.work_area_pen = QPen(QColor(0, 168, 230), 2, Qt.DashLine)
        
        # Track if Ctrl key is pressed
        self.ctrl_pressed = False
        self.temporary_work_area = None  # To store the temporary work area
        
        # Connect to the annotation window's image load signal
        self.annotation_window.imageLoaded.connect(self.on_image_loaded)
        
        # Track current image path to detect image changes
        self.current_image_path = None
        
        # Add hover position tracking for preview
        self.hover_pos = None
    
    def activate(self):
        """Activate the work area tool and set the appropriate cursor."""
        self.active = True
        self.annotation_window.viewport().setCursor(self.cursor)
        self.load_work_areas()

    def deactivate(self):
        """Deactivate the work area tool and clean up."""
        if self.drawing:
            self.cancel_drawing()
            
        # Remove all work area graphics from the scene without clearing the raster data
        self.clear_work_area_graphics()
        self.annotation_window.viewport().setCursor(self.default_cursor)
        self.active = False
        
        # Call parent deactivate to ensure crosshair is properly cleared
        super().deactivate()
        
    def clear_work_area_graphics(self):
        """Remove all work area graphics from the scene without clearing the data in the raster."""
        # Create a copy to safely iterate
        for work_area in self.work_areas:
            # Deanimate to unregister from animation manager
            work_area.deanimate()
            # Then handle the main graphics item and its children
            if work_area.graphics_item and work_area.graphics_item.scene():
                try:
                    # Remove from scene but keep the object
                    self.annotation_window.scene.removeItem(work_area.graphics_item)
                except RuntimeError as e:
                    print(f"Error removing graphics item: {e}")
                
            # Always clear references, even if scene removal fails
            work_area.graphics_item = None
            work_area.remove_button = None
        
        # Clear our internal list since we removed the graphics items
        # This doesn't remove them from the raster
        self.work_areas = []
        
    def mousePressEvent(self, event: QMouseEvent):
        """Handle mouse press events to start or finish drawing a work area."""
        if not self.annotation_window.active_image or not self.annotation_window.cursorInWindow(event.pos()):
            return
            
        # If Ctrl is pressed, don't start drawing - user is likely trying to delete work areas
        if self.ctrl_pressed:
            return
            
        if event.button() == Qt.LeftButton:
            scene_pos = self.annotation_window.mapToScene(event.pos())
            
            if not self.drawing:
                # Start drawing a new work area
                self.start_drawing(scene_pos)
            else:
                # Finish drawing the current work area
                self.finish_drawing(scene_pos)
                
    def mouseMoveEvent(self, event: QMouseEvent):
        """Handle mouse move events to update the work area while drawing."""
        # Call parent implementation to handle crosshair
        super().mouseMoveEvent(event)
        
        # Continue with tool-specific behavior
        scene_pos = self.annotation_window.mapToScene(event.pos())
        self.hover_pos = scene_pos  # Track hover position for spacebar confirmation
        
        if not self.drawing or not self.annotation_window.active_image:
            return
            
        self.update_drawing(scene_pos)
        
    def mouseReleaseEvent(self, event: QMouseEvent):
        """Handle mouse release events."""
        # For click-drag-click interactions, we don't need to handle release
        pass
        
    def keyPressEvent(self, event):
        """Handle key press events for work area tool operations."""
        modifiers = event.modifiers()
        key = event.key()

        # Confirm current drawing with spacebar
        if key == Qt.Key_Space and self.drawing and self.hover_pos:
            self.finish_drawing(self.hover_pos)
            return

        # Ctrl+Alt for temporary work area
        if (modifiers & Qt.ControlModifier) and (modifiers & Qt.AltModifier):
            if not self.temporary_work_area:
                self.temporary_work_area = self._create_temporary_work_area()
                if self.temporary_work_area:
                    self.save_work_area(self.temporary_work_area)  # Save to raster
            return

        # Ctrl+Shift to show remove buttons
        if (modifiers & Qt.ControlModifier) and (modifiers & Qt.ShiftModifier):
            self.ctrl_pressed = True
            if self.work_areas:  # Only show remove buttons if there are work areas exist
                self.update_remove_buttons_visibility(True)
                self.annotation_window.viewport().setCursor(Qt.PointingHandCursor)
            
            # Clear all work areas (Ctrl+Shift+Backspace)
            if key == Qt.Key_Backspace:
                if self.temporary_work_area:
                    self._remove_temporary_work_area()
                self.clear_work_areas()
            return

        # Ctrl+Space to create a work area from current view
        if key == Qt.Key_Space and self.annotation_window.active_image and not self.drawing:
            self.create_work_area_from_current_view()
            return

        # Cancel current drawing (Backspace or Escape - without modifiers)
        if (key == Qt.Key_Backspace or key == Qt.Key_Escape) and self.drawing and not (modifiers & Qt.ControlModifier):
            self.cancel_drawing()
            return

    def keyReleaseEvent(self, event):
        """Handle key release events."""
        modifiers = event.modifiers()
        
        # For Ctrl+Alt: remove temporary work area when either key is released
        if (event.key() == Qt.Key_Control or event.key() == Qt.Key_Alt) and not (
            (modifiers & Qt.ControlModifier) and (modifiers & Qt.AltModifier)
        ):
            if self.temporary_work_area is not None:
                self._remove_temporary_work_area()
        
        # For Ctrl+Shift: hide remove buttons when either key is released
        if (event.key() == Qt.Key_Control or event.key() == Qt.Key_Shift) and not (
            (modifiers & Qt.ControlModifier) and (modifiers & Qt.ShiftModifier)
        ):
            self.ctrl_pressed = False
            self.update_remove_buttons_visibility(False)
            self.annotation_window.viewport().setCursor(self.cursor)

    def _create_temporary_work_area(self):
        """Create a temporary work area from the current view."""
        viewport_rect = self.annotation_window.viewportToScene()
        constrained_viewport_rect = self.constrain_rect_to_image_bounds(viewport_rect)

        if constrained_viewport_rect.width() < 10 or constrained_viewport_rect.height() < 10:
            return None

        work_area = WorkArea.from_rect(constrained_viewport_rect, self.get_current_image_name())
        thickness = self.graphics_utility.get_workarea_thickness(self.annotation_window)
        work_area.create_graphics(self.annotation_window.scene, thickness)

        if work_area.graphics_item and work_area.graphics_item.scene():
            return work_area

        return None

    def _remove_temporary_work_area(self):
        """Remove the temporary work area from the scene and raster."""
        if self.temporary_work_area is None:
            return

        raster = self.get_current_raster()
        if raster:
            raster.remove_work_area(self.temporary_work_area)

        if self.temporary_work_area.graphics_item and self.temporary_work_area.graphics_item.scene():
            self.annotation_window.scene.removeItem(self.temporary_work_area.graphics_item)
            self.temporary_work_area.graphics_item = None

        self.temporary_work_area = None
        
    def start_drawing(self, pos):
        """Start drawing a work area at the given position."""
        self.drawing = True
        self.start_pos = pos
        
        # Create an initial rectangle item for visual feedback
        self.current_rect = QGraphicsRectItem(QRectF(pos.x(), pos.y(), 0, 0))
        
        # Create a dashed blue pen for the working area preview
        pen = QPen(QColor(0, 168, 230))
        pen.setStyle(Qt.DashLine)
        pen.setWidth(2)
        
        self.current_rect.setPen(pen)
        self.current_rect.setBrush(QBrush(QColor(0, 168, 230, 30)))  # Light blue transparent fill
        self.annotation_window.scene.addItem(self.current_rect)
        
    def update_drawing(self, pos):
        """Update the work area rectangle as the mouse moves."""
        if not self.current_rect:
            return
            
        # Create a normalized rectangle (handles drawing in any direction)
        rect = QRectF(self.start_pos, pos).normalized()
        
        # Constrain to image boundaries for visual feedback during drawing
        constrained_rect = self.constrain_rect_to_image_bounds(rect)
        
        self.current_rect.setRect(constrained_rect)
        self.hover_pos = pos  # Update hover position for key events
        
    def finish_drawing(self, pos):
        """Finish drawing the work area and add it to the list."""
        if not self.current_rect:
            self.cancel_drawing()
            return
            
        # Get the final rectangle
        rect = QRectF(self.start_pos, pos).normalized()
        
        # Ensure the rectangle stays within image boundaries
        rect = self.constrain_rect_to_image_bounds(rect)
        
        # Check if the work area is too small
        if rect.width() < 10 or rect.height() < 10:
            QMessageBox.warning(
                self.annotation_window,
                "Work Area Too Small",
                "Work area is too small. Please create a larger area."
            )
            self.cancel_drawing()
            return
            
        # Remove the temporary drawing rectangle
        if self.current_rect in self.annotation_window.scene.items():
            self.annotation_window.scene.removeItem(self.current_rect)
            
        # Create a WorkArea object from the rect
        work_area = WorkArea.from_rect(rect, self.get_current_image_name())
        
        # Set the animation manager reference
        work_area.set_animation_manager(self.annotation_window.animation_manager)
        
        # Create graphics using the WorkArea's own method
        thickness = self.graphics_utility.get_workarea_thickness(self.annotation_window)
        work_area.create_graphics(self.annotation_window.scene, thickness)
        work_area.animate()
        
        # Add close button but initially hidden unless Ctrl is pressed
        button_size = self.graphics_utility.get_handle_size(self.annotation_window) * 2
        work_area.create_remove_button(button_size, thickness)
        work_area.set_remove_button_visibility(self.ctrl_pressed)
        
        # Connect to remove signal
        work_area.removed.connect(self.on_work_area_removed)
        
        # Add to work areas list
        self.work_areas.append(work_area)
        
        # Store the work area in the raster
        self.save_work_area(work_area)
        
        # Reset state
        self.drawing = False
        self.start_pos = None
        self.current_rect = None
        
    def cancel_drawing(self):
        """Cancel the current work area drawing."""
        if self.current_rect:
            self.annotation_window.scene.removeItem(self.current_rect)
            self.current_rect = None
            
        self.drawing = False
        self.start_pos = None
        
    def create_work_area_from_current_view(self):
        """Create a work area based on the current viewport view."""
        viewport_rect = self.annotation_window.viewportToScene()
        
        # Ensure the rectangle stays within image boundaries
        viewport_rect = self.constrain_rect_to_image_bounds(viewport_rect)
        
        # Check if the constrained rectangle is too small
        if viewport_rect.width() < 10 or viewport_rect.height() < 10:
            QMessageBox.information(
                self.annotation_window,
                "Work Area Too Small",
                "The visible area within the image is too small to create a work area."
            )
            return
            
        # Create a WorkArea object from the viewport rect
        work_area = WorkArea.from_rect(viewport_rect, self.get_current_image_name())
        
        # Add work area graphics to the scene in one step
        thickness = self.graphics_utility.get_workarea_thickness(self.annotation_window)
        button_size = self.graphics_utility.get_handle_size(self.annotation_window) * 2
        
        if work_area.add_to_scene(self.annotation_window.scene, thickness, button_size):
            # Set initial button visibility based on Ctrl state
            work_area.set_remove_button_visibility(self.ctrl_pressed)
            
            # Connect to remove signal
            work_area.removed.connect(self.on_work_area_removed)
            
            # Add to work areas list
            self.work_areas.append(work_area)
            
            # Store the work area in the raster
            self.save_work_area(work_area)
        else:
            print(f"Warning: Failed to add work area to scene: {work_area.to_dict()}")
            
    def constrain_rect_to_image_bounds(self, rect):
        """Constrain a rectangle to stay within the image boundaries."""
        # Get image boundaries
        if self.annotation_window.pixmap_image:
            image_rect = QGraphicsPixmapItem(self.annotation_window.pixmap_image).boundingRect()
            
            # Create a copy of the input rect to avoid modifying the original
            constrained_rect = QRectF(rect)
            
            # If rect is completely outside image, return a minimal valid rect at image edge
            if (rect.right() < 0 or rect.bottom() < 0 or 
                rect.left() > image_rect.width() or rect.top() > image_rect.height()):
                # Return a small rect at the closest corner
                x = max(0, min(rect.x(), image_rect.width() - 10))
                y = max(0, min(rect.y(), image_rect.height() - 10))
                return QRectF(x, y, 10, 10)
            
            # Constrain top-left corner
            if constrained_rect.left() < 0:
                constrained_rect.setLeft(0)
            if constrained_rect.top() < 0:
                constrained_rect.setTop(0)
                
            # Constrain bottom-right corner
            if constrained_rect.right() > image_rect.width():
                constrained_rect.setRight(image_rect.width())
            if constrained_rect.bottom() > image_rect.height():
                constrained_rect.setBottom(image_rect.height())
                
            return constrained_rect
        
        return rect
        
    def update_remove_buttons_visibility(self, visible):
        """Update the visibility of all remove buttons based on Ctrl key state."""
        for work_area in self.work_areas:
            work_area.set_remove_button_visibility(visible)
        
    def on_work_area_removed(self, work_area):
        """Handle when a work area is removed."""
        # Remove from work areas list
        if work_area in self.work_areas:
            self.work_areas.remove(work_area)
            
        # Also remove from the raster's work areas list
        raster = self.get_current_raster()
        if raster:
            raster.remove_work_area(work_area)
        
    def get_current_image_name(self):
        """Get the name of the current image being displayed."""
        if not self.annotation_window.current_image_path:
            return None
            
        return self.annotation_window.current_image_path
        
    def get_current_raster(self):
        """Get the raster object for the current image."""
        image_path = self.get_current_image_name()
        if not image_path:
            return None
            
        return self.annotation_window.main_window.image_window.raster_manager.get_raster(image_path)
        
    def on_image_loaded(self, width, height):
        """Handle when a new image is loaded."""
        # Get current image path
        new_image_path = self.get_current_image_name()
        
        # Check if the image has actually changed
        if new_image_path != self.current_image_path:
            # Update our tracking variable
            self.current_image_path = new_image_path
            
            # Always reload work areas when the image changes, regardless of whether this tool is active
            if self.annotation_window.selected_tool == "work_area":
                self.load_work_areas()
        
    def check_and_reload_work_areas(self):
        """
        Check if we need to reload work areas - this can be called when the tool becomes
        visible or when we need to ensure graphics are displayed.
        """
        # If this is the active tool, always reload work areas
        if self.annotation_window.selected_tool == "work_area":
            self.load_work_areas()
            
    def load_work_areas(self):
        """Load existing work areas for the current image."""
        # Remove existing work area graphics from the scene first
        self.clear_work_area_graphics()
        
        # Get the raster for the current image
        raster = self.get_current_raster()
        if not raster:
            return
            
        # Get work areas from the raster's work_areas list
        stored_work_areas = raster.get_work_areas()
        image_path = self.get_current_image_name()
        
        if not stored_work_areas:
            return
            
        # Create graphics for each stored work area
        for work_area in stored_work_areas:
            # If the work area already has an image path that matches the current image
            # and a valid rect, use it directly
            if work_area.image_path == image_path and work_area.is_valid():
                # Ensure work area has no existing graphics references
                # This is important when reloading previously viewed images
                work_area.graphics_item = None
                work_area.remove_button = None
                
                # Set the animation manager reference
                work_area.set_animation_manager(self.annotation_window.animation_manager)
                
                # Add work area graphics to the scene
                thickness = self.graphics_utility.get_workarea_thickness(self.annotation_window)
                button_size = self.graphics_utility.get_handle_size(self.annotation_window) * 2
                
                if work_area.add_to_scene(self.annotation_window.scene, thickness, button_size):
                    # Animate the work area
                    work_area.animate()
                    # Set initial button visibility based on Ctrl state
                    work_area.set_remove_button_visibility(self.ctrl_pressed)
                    
                    # Connect to remove signal
                    # Disconnect first to avoid duplicate connections
                    try:
                        work_area.removed.disconnect(self.on_work_area_removed)
                    except TypeError:
                        # Ignore if not connected
                        pass
                    work_area.removed.connect(self.on_work_area_removed)
                    
                    # Add to work areas list
                    self.work_areas.append(work_area)
                else:
                    print(f"Warning: Failed to add work area to scene: {work_area.to_dict()}")
                    # Debug information
                    print(f"  - Graphics item: {work_area.graphics_item}")
                    print(f"  - Remove button: {work_area.remove_button}")
            
    def save_work_area(self, work_area):
        """Save the work area to the current raster."""
        raster = self.get_current_raster()
        if not raster:
            return
            
        # Add to raster's work_areas list
        raster.add_work_area(work_area)
            
    def clear_work_areas(self):
        """Clear all work areas completely - removing both graphics and data from the raster."""
        # Get the current raster first
        raster = self.get_current_raster()
        
        # Create a copy of the list to safely iterate and remove
        work_areas_copy = self.work_areas.copy()
        
        # Clear the list first to avoid on_work_area_removed removing during iteration
        self.work_areas = []
        
        # Remove each work area's graphics item from the scene
        for work_area in work_areas_copy:
            # If the graphics item exists and is in the scene, it will automatically
            # remove all its children (including the remove_button) when removed
            if work_area.graphics_item and work_area.graphics_item.scene():
                self.annotation_window.scene.removeItem(work_area.graphics_item)
                work_area.graphics_item = None
                # We don't need to explicitly remove the button as it was a child item
                # Just make sure we clear our reference to it
                work_area.remove_button = None
                
        # Also clear all work areas from the raster data
        if raster:
            raster.clear_work_areas()
        
    def update_cursor_annotation(self, scene_pos=None):
        """Method required by tool interface but not used for work area tool."""
        pass
        
    def clear_cursor_annotation(self):
        """Method required by tool interface but not used for work area tool."""
        pass


import warnings

warnings.filterwarnings("ignore")

import random
import math

from PyQt5.QtGui import QPainter, QColor, QFont, QPen, QBrush
from PyQt5.QtCore import Qt, pyqtSignal, QRect, QTimer, QPoint
from PyQt5.QtWidgets import (QMainWindow, QApplication, QDesktopWidget, QWidget, 
                             QMessageBox, QDialog, QVBoxLayout, QHBoxLayout, QPushButton, 
                             QLabel, QButtonGroup)

from coralnet_toolbox.Icons import get_icon


# ----------------------------------------------------------------------------------------------------------------------
# Constants / Configurations
# ----------------------------------------------------------------------------------------------------------------------


# Define base game constants
BASE_WIDTH = 700
BASE_HEIGHT = 700
BASE_PADDLE_WIDTH = 75
PADDLE_HEIGHT = 15
PADDLE_SPEED = 15
BASE_BALL_DIAMETER = 15
BASE_GAME_SPEED = 5  # Lower is faster


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class BallTrail:
    """Represents a trailing effect behind the ball."""
    def __init__(self, x, y, size):
        self.x = x
        self.y = y
        self.size = size
        self.life = 30  # Increased from 15 to 30 for a longer tail
        self.max_life = 30
        
    def update(self):
        """Update trail position and properties."""
        self.life -= 1
        self.size = max(1, self.size - 0.15)  # Slower shrink for longer visible trail
        return self.life > 0
        
    def draw(self, painter):
        """Draw the trail segment."""
        if self.life > 0:
            # Fade out over time
            alpha = int(200 * (self.life / self.max_life))
            color = QColor(211, 211, 211)  # Light gray trail
            color.setAlpha(alpha)
            painter.setBrush(QBrush(color))
            painter.setPen(Qt.NoPen)
            painter.drawEllipse(int(self.x), int(self.y), int(self.size), int(self.size))


class Particle:
    """Represents a visual particle for brick destruction effects."""
    def __init__(self, x, y, color, velocity_x=None, velocity_y=None):
        self.x = x
        self.y = y
        self.color = color
        self.velocity_x = velocity_x if velocity_x is not None else random.uniform(-3, 3)
        self.velocity_y = velocity_y if velocity_y is not None else random.uniform(-5, -1)
        self.life = 30  # Frames to live
        self.size = random.randint(2, 5)
        self.gravity = 0.2
        
    def update(self):
        """Update particle position and properties."""
        self.x += self.velocity_x
        self.y += self.velocity_y
        self.velocity_y += self.gravity
        self.life -= 1
        self.size = max(1, self.size - 0.1)
        return self.life > 0
        
    def draw(self, painter):
        """Draw the particle."""
        if self.life > 0:
            # Fade out over time
            alpha = int(255 * (self.life / 30))
            color = QColor(self.color)
            color.setAlpha(alpha)
            painter.setBrush(QBrush(color))
            painter.setPen(Qt.NoPen)
            painter.drawEllipse(int(self.x), int(self.y), int(self.size), int(self.size))


class PowerUp:
    """Represents a power-up that falls from destroyed bricks."""
    def __init__(self, x, y, power_type):
        self.rect = QRect(x, y, 20, 20)
        self.power_type = power_type
        self.velocity_y = 2
        self.colors = {
            'multiball': QColor(255, 0, 255),       # Magenta
            'big_paddle': QColor(0, 255, 0),        # Green
            'small_paddle': QColor(255, 165, 0),    # Orange
            'laser': QColor(255, 0, 0),             # Red
            'slow_ball': QColor(0, 255, 255),       # Cyan (brighter than blue)
            'fast_ball': QColor(255, 255, 255),     # White
            'reverse_paddle': QColor(128, 0, 128),  # Purple
            'extra_life': QColor(255, 20, 147)      # Deep Pink
        }
        
    def update(self):
        """Update power-up position."""
        self.rect.translate(0, self.velocity_y)
        
    def draw(self, painter):
        """Draw the power-up."""
        painter.setBrush(QBrush(self.colors[self.power_type]))
        painter.setPen(QPen(QColor(0, 0, 0), 2))
        painter.drawEllipse(self.rect)
        
        # Draw power-up symbol
        painter.setPen(QPen(QColor(0, 0, 0), 2))
        painter.setFont(QFont('Arial', 8, QFont.Bold))
        symbols = {
            'multiball': 'M',
            'big_paddle': '+',
            'small_paddle': '-',
            'laser': 'L',
            'slow_ball': 'SL',
            'fast_ball': 'F',
            'reverse_paddle': 'R',
            'extra_life': '♥'
        }
        symbol = symbols.get(self.power_type, '?')
        painter.drawText(self.rect, Qt.AlignCenter, symbol)


class Laser:
    """Represents a laser shot from the paddle."""
    def __init__(self, x, y):
        self.rect = QRect(x, y, 3, 10)
        self.velocity_y = -8
        
    def update(self):
        """Update laser position."""
        self.rect.translate(0, self.velocity_y)
        
    def draw(self, painter):
        """Draw the laser."""
        painter.setBrush(QBrush(QColor(255, 0, 0)))
        painter.setPen(Qt.NoPen)
        painter.drawRect(self.rect)


class HighScoreManager:
    """Manages high scores storage and retrieval."""
    def __init__(self):
        self.high_scores = {"Easy": [], "Medium": [], "Hard": [], "Insane": []}
        
    def add_score(self, difficulty, score):
        """Add a new score and return if it's a high score."""
        if difficulty not in self.high_scores:
            self.high_scores[difficulty] = []
            
        self.high_scores[difficulty].append(score)
        self.high_scores[difficulty].sort(reverse=True)
        self.high_scores[difficulty] = self.high_scores[difficulty][:10]  # Keep top 10
        
        return score in self.high_scores[difficulty][:3]  # Top 3 is considered "high score"
        
    def get_high_score(self, difficulty):
        """Get the highest score for a difficulty."""
        if difficulty in self.high_scores and self.high_scores[difficulty]:
            return self.high_scores[difficulty][0]
        return 0


class Board(QWidget):
    """
    The main game board where all the action happens.
    It handles drawing, game logic, and user input.
    """
    # Define base game constants
    WIDTH = BASE_WIDTH
    HEIGHT = BASE_HEIGHT
    PADDLE_WIDTH = BASE_PADDLE_WIDTH
    PADDLE_HEIGHT = PADDLE_HEIGHT
    PADDLE_SPEED = PADDLE_SPEED
    BALL_DIAMETER = BASE_BALL_DIAMETER
    GAME_SPEED = BASE_GAME_SPEED

    def __init__(self, parent, difficulty):
        super().__init__(parent)
        self.parent_window = parent
        self.difficulty = difficulty
        self.high_score_manager = HighScoreManager()
        self.setup_difficulty_parameters()
        self.initGame()

    def setup_difficulty_parameters(self):
        """Set game parameters based on difficulty level."""
        difficulty_settings = {
            "Easy": {
                "paddle_width_multiplier": 1.5,
                "ball_speed_multiplier": 0.7,
                "game_speed": 10,  # Slower
                "brick_rows": 6,
                "ball_diameter_multiplier": 1.2,
                "score_multiplier": 1.0,
                "width_multiplier": 0.8,  # Smaller play area
                "height_multiplier": 0.8
            },
            "Medium": {
                "paddle_width_multiplier": 1.0,
                "ball_speed_multiplier": 1.0,
                "game_speed": 8,  # Normal
                "brick_rows": 8,
                "ball_diameter_multiplier": 1.0,
                "score_multiplier": 1.5,
                "width_multiplier": 1.0,  # Normal play area
                "height_multiplier": 1.0
            },
            "Hard": {
                "paddle_width_multiplier": 0.7,
                "ball_speed_multiplier": 1.4,
                "game_speed": 6,  # Faster
                "brick_rows": 10,
                "ball_diameter_multiplier": 0.8,
                "score_multiplier": 2.0,
                "width_multiplier": 1.2,  # Larger play area
                "height_multiplier": 1.2
            },
            "Insane": {
                "paddle_width_multiplier": 0.5,
                "ball_speed_multiplier": 2.0,
                "game_speed": 4,  # Very fast
                "brick_rows": 12,
                "ball_diameter_multiplier": 0.6,
                "score_multiplier": 3.0,
                "width_multiplier": 1.4,  # Much larger play area
                "height_multiplier": 1.4
            }
        }
        
        settings = difficulty_settings[self.difficulty]
        
        # Apply difficulty settings including dynamic width/height
        self.WIDTH = int(BASE_WIDTH * settings["width_multiplier"])
        self.HEIGHT = int(BASE_HEIGHT * settings["height_multiplier"])
        self.PADDLE_WIDTH = int(BASE_PADDLE_WIDTH * settings["paddle_width_multiplier"])
        self.BALL_DIAMETER = int(BASE_BALL_DIAMETER * settings["ball_diameter_multiplier"])
        self.GAME_SPEED = settings["game_speed"]
        self.BRICK_ROWS = settings["brick_rows"]
        self.ball_speed_multiplier = settings["ball_speed_multiplier"]
        self.score_multiplier = settings["score_multiplier"]
        
        # Adjust paddle speed based on difficulty (smaller paddles move faster for compensation)
        self.PADDLE_SPEED = int(PADDLE_SPEED * (2 - settings["paddle_width_multiplier"]))

    def initGame(self):
        """Initializes all game variables and objects."""
        # Game state flags
        self.isStarted = False
        self.isPaused = False
        
        # Game progression
        self.current_level = 1
        self.score = 0
        self.lives = 3
        self.combo_count = 0
        
        # Ball speed progression
        self.speed_increase_factor = 1.0  # 5% speed increase per collision
        self.max_speed_multiplier = 5     # Maximum speed multiplier
        
        # Paddle setup
        self.original_paddle_width = self.PADDLE_WIDTH
        self.paddle = QRect((self.WIDTH - self.PADDLE_WIDTH) // 2, 
                            self.HEIGHT - 50, 
                            self.PADDLE_WIDTH, self.PADDLE_HEIGHT)

        # Ball setup
        self.balls = []
        self.resetBall()
        
        # Sticky paddle tracking
        self.sticky_balls = []  # List of balls that are stuck to paddle
        
        # Power-up system
        self.power_ups = []
        self.active_power_ups = {}
        self.power_up_timers = {}
        self.lasers = []
        
        # Visual effects
        self.particles = []
        self.trails = []  # List of active trails

        # Bricks setup
        self.initBricks()

        # Timer for game loop
        self.timer = QTimer(self)
        
        # Power-up effect timer
        self.effect_timer = QTimer()
        self.effect_timer.timeout.connect(self.update_power_up_effects)
        self.effect_timer.start(100)  # Update every 100ms
        
        # Enable keyboard focus
        self.setFocusPolicy(Qt.StrongFocus)

    def initBricks(self):
        """Creates the grid of bricks based on difficulty and level."""
        self.bricks = []
        
        # Dynamic brick sizing based on screen dimensions
        BRICK_WIDTH = max(40, min(60, self.WIDTH // 15))  # Scale brick width
        BRICK_HEIGHT = max(15, min(25, self.HEIGHT // 35))  # Scale brick height
        BRICK_COLS = max(8, min(14, (self.WIDTH - 40) // (BRICK_WIDTH + 5)))  # Dynamic columns
        PADDING = max(3, self.WIDTH // 160)  # Scale padding
        TOP_OFFSET = max(85, self.HEIGHT // 10)  # Scale top offset (below UI bar)

        # Base colors
        base_colors = [QColor(0, 255, 0), QColor(0, 0, 255), QColor(255, 255, 0),
                       QColor(255, 165, 0), QColor(255, 0, 0), QColor(128, 0, 128),
                       QColor(255, 192, 203), QColor(0, 255, 255), QColor(255, 20, 147), QColor(50, 205, 50)]

        # Adjust brick layout based on level and available space
        base_rows = self.BRICK_ROWS + (self.current_level - 1) // 2  # More rows each 2 levels
        max_rows = min(base_rows, (self.HEIGHT - TOP_OFFSET - 150) // (BRICK_HEIGHT + PADDING))  # Ensure bricks fit
        rows = max(3, max_rows)  # Minimum 3 rows
        
        # Center the brick layout horizontally
        total_brick_width = BRICK_COLS * BRICK_WIDTH + (BRICK_COLS - 1) * PADDING
        start_x = (self.WIDTH - total_brick_width) // 2
        
        for row in range(rows):
            for col in range(BRICK_COLS):
                x = start_x + col * (BRICK_WIDTH + PADDING)
                y = row * (BRICK_HEIGHT + PADDING) + TOP_OFFSET
                color = base_colors[row % len(base_colors)]
                
                # Special bricks for higher levels
                brick_type = 'normal'
                brick_hits = 1
                
                if self.current_level >= 3 and random.random() < 0.1:  # 10% chance of strong bricks
                    brick_type = 'strong'
                    brick_hits = 2
                    color = color.darker(150)
                
                self.bricks.append({
                    'rect': QRect(x, y, BRICK_WIDTH, BRICK_HEIGHT), 
                    'color': color,
                    'type': brick_type,
                    'hits': brick_hits,
                    'points': (row + 1) * 10 * self.score_multiplier
                })

    def resetBall(self):
        """Resets the ball to its starting position and state."""
        # Clear existing balls first
        self.balls = []
        ball = QRect(self.paddle.center().x() - self.BALL_DIAMETER // 2,
                     self.paddle.top() - self.BALL_DIAMETER,
                     self.BALL_DIAMETER, self.BALL_DIAMETER)
        
        # Reset to base speed, ignoring any accumulated speed increases
        base_speed = 1
        self.balls.append({
            'rect': ball,
            'xDir': int(base_speed * self.ball_speed_multiplier),
            'yDir': int(-base_speed * self.ball_speed_multiplier)
        })
        
        # Reset the speed increase factor for the new life
        self.speed_increase_factor = 1.0

    def create_particles(self, x, y, color, count=8):
        """Create particle effects at the given position."""
        for _ in range(count):
            particle = Particle(x, y, color)
            self.particles.append(particle)

    def spawn_power_up(self, x, y):
        """Randomly spawn a power-up at the given position."""
        if random.random() < 0.25:  # 25% chance (increased from 15%)
            power_types = ['multiball', 
                           'big_paddle', 
                           'small_paddle', 
                           'laser', 
                           'slow_ball', 
                           'fast_ball',
                           'reverse_paddle']
            
            if self.current_level >= 3:
                power_types.append('extra_life')
                
            power_type = random.choice(power_types)
            power_up = PowerUp(x, y, power_type)
            self.power_ups.append(power_up)

    def apply_power_up(self, power_type):
        """Apply the effect of a power-up."""
        if power_type == 'multiball':
            # Create additional balls
            for _ in range(2):
                if self.balls:
                    original_ball = self.balls[0]
                    new_ball = {
                        'rect': QRect(original_ball['rect']),
                        'xDir': original_ball['xDir'] + random.randint(-2, 2),
                        'yDir': original_ball['yDir']
                    }
                    self.balls.append(new_ball)
                    
        elif power_type == 'big_paddle':
            self.paddle = QRect(self.paddle.x(), 
                                self.paddle.y(), 
                                min(self.original_paddle_width * 2, 150), 
                                self.PADDLE_HEIGHT)
            
            self.active_power_ups['big_paddle'] = True
            self.power_up_timers['big_paddle'] = 500  # 5 seconds at 100ms intervals
            
        elif power_type == 'small_paddle':
            self.paddle = QRect(self.paddle.x(), 
                                self.paddle.y(), 
                                max(self.original_paddle_width // 2, 30), 
                                self.PADDLE_HEIGHT)
            
            self.active_power_ups['small_paddle'] = True
            self.power_up_timers['small_paddle'] = 500
            
        elif power_type == 'laser':
            self.active_power_ups['laser'] = True
            self.power_up_timers['laser'] = 1000
            
        elif power_type == 'slow_ball':
            # Store original speeds before slowing down
            if 'slow_ball' not in self.active_power_ups:
                self.original_ball_speeds = []
                for ball in self.balls:
                    self.original_ball_speeds.append({'xDir': ball['xDir'], 'yDir': ball['yDir']})
                    
            # Apply slow effect to all balls
            for ball in self.balls:
                new_x_dir = int(ball['xDir'] * 0.5)
                new_y_dir = int(ball['yDir'] * 0.5)
                
                # Ensure minimum movement to prevent ball from stopping
                if abs(new_x_dir) < 1:
                    new_x_dir = 1 if ball['xDir'] >= 0 else -1
                if abs(new_y_dir) < 1:
                    new_y_dir = 1 if ball['yDir'] >= 0 else -1
                    
                ball['xDir'] = new_x_dir
                ball['yDir'] = new_y_dir
                
            self.active_power_ups['slow_ball'] = True
            self.power_up_timers['slow_ball'] = 1000
            
        elif power_type == 'fast_ball':
            # Store original speeds before speeding up
            if 'fast_ball' not in self.active_power_ups:
                self.original_ball_speeds = []
                for ball in self.balls:
                    self.original_ball_speeds.append({'xDir': ball['xDir'], 'yDir': ball['yDir']})
                    
            # Apply fast effect to all balls
            for ball in self.balls:
                new_x_dir = int(ball['xDir'] * 2.0)  # Reduced from 3.0 to 2.0 for better gameplay
                new_y_dir = int(ball['yDir'] * 2.0)
                
                # Ensure reasonable maximum speed
                max_speed = int(6 * self.ball_speed_multiplier)  # Increased max for fast ball
                new_x_dir = max(-max_speed, min(max_speed, new_x_dir))
                new_y_dir = max(-max_speed, min(max_speed, new_y_dir))
                
                # Ensure minimum movement
                if abs(new_x_dir) < 2:  # Higher minimum for fast ball
                    new_x_dir = 2 if ball['xDir'] >= 0 else -2
                if abs(new_y_dir) < 2:
                    new_y_dir = 2 if ball['yDir'] >= 0 else -2
                    
                ball['xDir'] = new_x_dir
                ball['yDir'] = new_y_dir
                
            self.active_power_ups['fast_ball'] = True
            self.power_up_timers['fast_ball'] = 500
            
        elif power_type == 'reverse_paddle':
            self.active_power_ups['reverse_paddle'] = True
            self.power_up_timers['reverse_paddle'] = 800  # 8 seconds
            
        elif power_type == 'extra_life':
            self.lives += 1
            
    def update_power_up_effects(self):
        """Update power-up timers and remove expired effects."""
        expired_effects = []
        
        for effect, timer in self.power_up_timers.items():
            self.power_up_timers[effect] -= 1
            if self.power_up_timers[effect] <= 0:
                expired_effects.append(effect)
                
        for effect in expired_effects:
            self.remove_power_up_effect(effect)
            
    def remove_power_up_effect(self, effect):
        """Remove a power-up effect."""
        if effect in self.active_power_ups:
            del self.active_power_ups[effect]
        if effect in self.power_up_timers:
            del self.power_up_timers[effect]
            
        if effect in ['big_paddle', 'small_paddle']:
            # Reset paddle to original size
            self.paddle = QRect(self.paddle.x(), self.paddle.y(), 
                                self.original_paddle_width, 
                                self.PADDLE_HEIGHT)
        elif effect in ['slow_ball', 'fast_ball']:
            # Restore original ball speeds
            if hasattr(self, 'original_ball_speeds') and self.original_ball_speeds:
                for i, ball in enumerate(self.balls):
                    if i < len(self.original_ball_speeds):
                        # Restore the original direction and speed magnitude
                        original = self.original_ball_speeds[i]
                        ball['xDir'] = original['xDir']
                        ball['yDir'] = original['yDir']
                        
                # Clear the stored speeds
                self.original_ball_speeds = []

    def paintEvent(self, event):
        """Handles all the drawing."""
        painter = QPainter(self)
        
        # Draw black background
        painter.fillRect(self.rect(), QColor(0, 0, 0))
        self.drawObjects(painter)

        # Draw breakout title animation
        if hasattr(self.parent_window, 'title_timer') and self.parent_window.title_timer > 0:
            title_font = QFont('Arial', 36, QFont.Bold)
            painter.setFont(title_font)

            # Create animated colors using bright game colors
            colors = [QColor(255, 0, 0), QColor(0, 255, 0), QColor(0, 255, 255), QColor(255, 255, 0)]
            color_index = (self.parent_window.title_timer // 10) % len(colors)
            painter.setPen(colors[color_index])

            title_text = "BREAKOUT GAME"
            text_rect = painter.fontMetrics().boundingRect(title_text)
            x = (self.width() - text_rect.width()) // 2
            y = self.height() // 2

            # Add glow effect to title
            glow_colors = [QColor(255, 0, 0, 150), QColor(0, 255, 0, 150)]
            glow_index = (self.parent_window.title_timer // 5) % len(glow_colors)
            painter.setPen(glow_colors[glow_index])
            painter.drawText(x - 2, y - 2, title_text)
            painter.drawText(x + 2, y - 2, title_text)
            painter.drawText(x - 2, y + 2, title_text)
            painter.drawText(x + 2, y + 2, title_text)

            painter.setPen(colors[color_index])
            painter.drawText(x, y, title_text)

    def drawObjects(self, painter):
        """Draws all game objects."""
        # Draw UI background bar
        ui_height = 85
        painter.setBrush(QBrush(QColor(0, 0, 0)))  # Fully black
        painter.setPen(Qt.NoPen)
        painter.drawRect(0, 0, self.WIDTH, ui_height)

        # Set up fonts and colors
        title_font = QFont('Arial', 14, QFont.Bold)
        info_font = QFont('Arial', 11)
        small_font = QFont('Arial', 9)

        # Left section - Game stats
        painter.setPen(QColor(255, 255, 255))  # White text
        painter.setFont(info_font)
        painter.drawText(15, 25, f"Level: {self.current_level}")
        painter.drawText(15, 45, f"Score: {self.score:,}")
        painter.drawText(15, 65, f"Lives: {self.lives}")

        # Center section - Difficulty info
        center_x = self.WIDTH // 2
        painter.setPen(QColor(255, 255, 255))  # White text
        painter.setFont(title_font)
        painter.drawText(center_x - 50, 20, f"{self.difficulty} Mode")

        painter.setFont(info_font)
        painter.setPen(QColor(255, 255, 255))  # White text
        high_score = self.high_score_manager.get_high_score(self.difficulty)
        painter.drawText(center_x - 50, 40, f"High Score: {high_score:,}")

        # Right section - Active power-ups
        right_x = self.WIDTH - 200
        if self.active_power_ups:
            painter.setPen(QColor(255, 255, 255))  # White text
            painter.setFont(title_font)
            painter.drawText(right_x, 20, "Power-ups:")

            painter.setFont(small_font)
            y_offset = 35
            power_up_colors = {
                'multiball': QColor(255, 0, 255),       # Magenta
                'big_paddle': QColor(0, 255, 0),        # Green
                'small_paddle': QColor(255, 165, 0),    # Orange
                'laser': QColor(255, 0, 0),             # Red
                'slow_ball': QColor(0, 255, 255),       # Cyan (brighter than blue)
                'fast_ball': QColor(255, 255, 255),     # White
                'reverse_paddle': QColor(128, 0, 128),  # Purple
                'sticky_paddle': QColor(255, 255, 0)    # Yellow
            }

            # Show at most 3 power-ups to fit in the UI bar
            power_up_list = list(self.active_power_ups.keys())[:3]
            remaining_count = len(self.active_power_ups) - len(power_up_list)

            for power_up in power_up_list:
                # Draw colored indicator circle
                color = power_up_colors.get(power_up, QColor(200, 200, 200))
                painter.setBrush(color)
                painter.setPen(Qt.NoPen)
                painter.drawEllipse(right_x - 12, y_offset - 6, 8, 8)

                # Draw power-up name
                painter.setPen(QColor(255, 255, 255))  # White text
                display_name = power_up.replace('_', ' ').title()
                painter.drawText(right_x + 5, y_offset, display_name)
                y_offset += 15

            # If there are more power-ups, show a count
            if remaining_count > 0:
                painter.setPen(QColor(255, 255, 255))  # White text
                painter.drawText(right_x + 5, y_offset, f"+{remaining_count} more")
        else:
            painter.setPen(QColor(255, 255, 255))  # White text
            painter.setFont(info_font)
            painter.drawText(right_x, 35, "No active power-ups")
        
        # Draw Paddle
        paddle_color = QColor(0, 180, 255)
        if 'laser' in self.active_power_ups:
            paddle_color = QColor(255, 0, 0)  # Red for laser mode
        elif 'sticky_paddle' in self.active_power_ups:
            paddle_color = QColor(255, 255, 0)  # Yellow for sticky
            
        painter.setBrush(paddle_color)
        painter.setPen(Qt.NoPen)
        painter.drawRect(self.paddle)

        # Draw Trails (behind balls)
        for trail in self.trails:
            trail.draw(painter)
            
        # Draw Balls
        for ball in self.balls:
            painter.setBrush(QColor(255, 255, 255))
            painter.drawEllipse(ball['rect'])

        # Draw Bricks
        for brick in self.bricks:
            painter.setBrush(brick['color'])
            painter.setPen(QPen(QColor(30, 30, 30), 2))
            painter.drawRect(brick['rect'])
            
            # Draw hit indicator for strong bricks
            if brick['type'] == 'strong' and brick['hits'] > 1:
                painter.setPen(QPen(QColor(255, 255, 255), 2))
                painter.setFont(QFont('Arial', 8, QFont.Bold))
                painter.drawText(brick['rect'], Qt.AlignCenter, str(brick['hits']))
                
        # Draw Power-ups
        for power_up in self.power_ups:
            power_up.draw(painter)
            
        # Draw Lasers
        for laser in self.lasers:
            laser.draw(painter)
            
        # Draw Particles
        for particle in self.particles:
            particle.draw(painter)
            
        # Draw start/pause message
        if not self.isStarted:
            self.drawInfoText(painter, "Press SPACE to start")
        elif self.isPaused:
            self.drawInfoText(painter, "PAUSED - Press P to continue")

    def drawInfoText(self, painter, text):
        """Draws informational text on the screen."""
        painter.setPen(QColor(255, 255, 255))  # White text
        painter.setFont(QFont('Arial', 15))
        text_rect = painter.fontMetrics().boundingRect(text)
        x = (self.width() - text_rect.width()) // 2
        y = self.height() - 200
        painter.drawText(x, y, text)

    def keyPressEvent(self, event):
        """Handles key presses."""
        key = event.key()
        old_paddle_x = self.paddle.x()

        # Check if reverse paddle is active
        is_reversed = 'reverse_paddle' in self.active_power_ups

        if key == Qt.Key_A:
            # Move left normally, or right if reversed
            if is_reversed:
                if self.paddle.right() < self.WIDTH:
                    self.paddle.translate(self.PADDLE_SPEED, 0)
                    self.move_stuck_balls_with_paddle(old_paddle_x)
                    # Also move ball if game hasn't started
                    if not self.isStarted:
                        self.move_ball_with_paddle_before_start(old_paddle_x)
            else:
                if self.paddle.left() > 0:
                    self.paddle.translate(-self.PADDLE_SPEED, 0)
                    self.move_stuck_balls_with_paddle(old_paddle_x)
                    # Also move ball if game hasn't started
                    if not self.isStarted:
                        self.move_ball_with_paddle_before_start(old_paddle_x)
        elif key == Qt.Key_D:
            # Move right normally, or left if reversed
            if is_reversed:
                if self.paddle.left() > 0:
                    self.paddle.translate(-self.PADDLE_SPEED, 0)
                    self.move_stuck_balls_with_paddle(old_paddle_x)
                    # Also move ball if game hasn't started
                    if not self.isStarted:
                        self.move_ball_with_paddle_before_start(old_paddle_x)
            else:
                if self.paddle.right() < self.WIDTH:
                    self.paddle.translate(self.PADDLE_SPEED, 0)
                    self.move_stuck_balls_with_paddle(old_paddle_x)
                    # Also move ball if game hasn't started
                    if not self.isStarted:
                        self.move_ball_with_paddle_before_start(old_paddle_x)
        elif key == Qt.Key_Space:
            if not self.isStarted and not self.isPaused:
                self.startGame()
            elif self.isStarted and not self.isPaused:
                # Check if we have stuck balls (from sticky paddle)
                stuck_balls = [ball for ball in self.balls if ball['xDir'] == 0 and ball['yDir'] == 0]
                if stuck_balls:
                    # Release stuck balls
                    self.release_stuck_balls()
                elif 'laser' in self.active_power_ups:
                    # Shoot laser
                    laser_x = self.paddle.center().x() - 1
                    laser_y = self.paddle.top() - 10
                    laser = Laser(laser_x, laser_y)
                    self.lasers.append(laser)
        elif key == Qt.Key_P:
            self.pauseGame()
        
        self.update()  # Redraw after key press

    def move_ball_with_paddle_before_start(self, old_paddle_x):
        """Move the ball with the paddle when the game hasn't started yet."""
        if not self.isStarted and self.balls:
            paddle_movement = self.paddle.x() - old_paddle_x
            # Move the ball to stay centered on the paddle
            for ball in self.balls:
                ball['rect'].translate(paddle_movement, 0)
                
    def move_stuck_balls_with_paddle(self, old_paddle_x):
        """Move any stuck balls along with the paddle movement."""
        if 'sticky_paddle' not in self.active_power_ups:
            return
            
        paddle_movement = self.paddle.x() - old_paddle_x
        
        for ball in self.balls:
            # Only move balls that are stuck (velocity is 0)
            if ball['xDir'] == 0 and ball['yDir'] == 0:
                ball['rect'].translate(paddle_movement, 0)
                
    def release_stuck_balls(self):
        """Release all stuck balls from the sticky paddle."""
        for ball in self.balls:
            if ball['xDir'] == 0 and ball['yDir'] == 0:
                # Calculate release angle based on ball position relative to paddle center
                paddle_center = self.paddle.center().x()
                ball_center = ball['rect'].center().x()
                
                # Determine angle based on position on paddle
                relative_position = (ball_center - paddle_center) / (self.paddle.width() / 2)
                # Clamp to reasonable range
                relative_position = max(-1, min(1, relative_position))
                
                # Convert to launch angle (between -60 and 60 degrees)
                launch_angle = relative_position * 60  # degrees
                
                # Set ball velocity
                base_speed = max(2, int(self.ball_speed_multiplier))
                ball['xDir'] = int(base_speed * math.sin(math.radians(launch_angle)))
                ball['yDir'] = int(-base_speed * math.cos(math.radians(launch_angle)))
                
                # Ensure minimum movement
                if abs(ball['xDir']) < 1:
                    ball['xDir'] = 1 if ball['xDir'] >= 0 else -1
                if abs(ball['yDir']) < 1:
                    ball['yDir'] = -1

    def startGame(self):
        """Starts the game loop."""
        if not self.isStarted:
            # Stop and disconnect any existing timer before starting
            if self.timer.isActive():
                self.timer.stop()
            
            # Disconnect all previous connections to prevent multiple connections
            try:
                self.timer.timeout.disconnect()
            except TypeError:
                # No connections to disconnect
                pass
                
            self.resetBall()
            self.isStarted = True
            self.timer.timeout.connect(self.update_game)
            self.timer.start(self.GAME_SPEED)

    def pauseGame(self):
        """Pauses or unpauses the game."""
        if self.isStarted:
            self.isPaused = not self.isPaused
            if self.isPaused:
                self.timer.stop()
            else:
                self.timer.start(self.GAME_SPEED)

    def update_game(self):
        """The main game loop, called by the timer."""
        if hasattr(self.parent_window, 'title_timer') and self.parent_window.title_timer > 0:
            self.parent_window.title_timer -= 1
            self.update()
            return

        self.moveBalls()
        self.updatePowerUps()
        self.updateLasers()
        self.updateParticles()
        self.updateTrails()
        self.checkCollision()
        self.update()

    def moveBalls(self):
        """Moves all balls according to their direction vectors."""
        if self.isStarted and not self.isPaused:
            for ball in self.balls:
                # Create trail at current position before moving
                ball_center = ball['rect'].center()
                trail = BallTrail(ball_center.x() - self.BALL_DIAMETER // 4, 
                                  ball_center.y() - self.BALL_DIAMETER // 4, 
                                  self.BALL_DIAMETER // 2)
                self.trails.append(trail)
                
                # Move the ball
                ball['rect'].translate(ball['xDir'], ball['yDir'])

    def updatePowerUps(self):
        """Update power-up positions and remove those that fall off screen."""
        for power_up in self.power_ups[:]:
            power_up.update()
            if power_up.rect.top() > self.HEIGHT:
                self.power_ups.remove(power_up)
                
    def updateLasers(self):
        """Update laser positions and remove those that go off screen."""
        for laser in self.lasers[:]:
            laser.update()
            if laser.rect.bottom() < 0:
                self.lasers.remove(laser)
                
    def updateParticles(self):
        """Update particle effects."""
        for particle in self.particles[:]:
            if not particle.update():
                self.particles.remove(particle)

    def updateTrails(self):
        """Update trail effects."""
        for trail in self.trails[:]:
            if not trail.update():
                self.trails.remove(trail)

    def checkCollision(self):
        """Checks for and handles collisions with comprehensive safeguards."""
        balls_to_remove = []
        
        for i, ball in enumerate(self.balls):
            # Enforce strict boundary safeguards to prevent getting stuck
            self.enforce_ball_boundaries(ball)
            
            # Ball hits bottom wall -> Lose a life
            if ball['rect'].bottom() > self.HEIGHT:
                balls_to_remove.append(i)
                continue

            # Ball hits walls with enhanced collision detection
            wall_collision = False
            
            # Left wall collision
            if ball['rect'].left() <= 0:
                ball['rect'].moveLeft(1)  # Move ball away from wall
                ball['xDir'] = abs(ball['xDir'])  # Ensure positive direction
                wall_collision = True
                
            # Right wall collision  
            elif ball['rect'].right() >= self.WIDTH:
                ball['rect'].moveRight(self.WIDTH - 1)  # Move ball away from wall
                ball['xDir'] = -abs(ball['xDir'])  # Ensure negative direction
                wall_collision = True
                
            # Top wall collision
            if ball['rect'].top() <= 0:
                ball['rect'].moveTop(1)  # Move ball away from wall
                ball['yDir'] = abs(ball['yDir'])  # Ensure positive direction (downward)
                wall_collision = True
                
            # If we had a wall collision, increase speed and ensure minimum speed
            if wall_collision:
                self.increase_ball_speed(ball)
                self.ensure_minimum_ball_speed(ball)

            # Ball hits paddle
            if ball['rect'].intersects(self.paddle) and ball['yDir'] > 0:
                # Move ball above paddle to prevent sticking
                ball['rect'].moveBottom(self.paddle.top() - 2)  # Extra margin
                
                if 'sticky_paddle' in self.active_power_ups:
                    # Sticky paddle - stop the ball
                    ball['xDir'] = 0
                    ball['yDir'] = 0
                else:
                    # Complex paddle collision for better control
                    paddle_center = self.paddle.center().x()
                    ball_center = ball['rect'].center().x()
                    
                    # Change horizontal direction based on where it hits the paddle
                    hit_factor = (ball_center - paddle_center) / (self.paddle.width() / 2)
                    max_speed = int(2 * self.ball_speed_multiplier)
                    ball['xDir'] = max(-max_speed, min(max_speed, int(hit_factor * max_speed)))
                    if ball['xDir'] == 0:
                        if hit_factor >= 0:
                            ball['xDir'] = int(self.ball_speed_multiplier)
                        else:
                            ball['xDir'] = int(-self.ball_speed_multiplier)
                    ball['yDir'] = int(-1 * self.ball_speed_multiplier)  # Always bounce up
                    
                    # Increase speed after paddle collision and ensure minimum speed
                    self.increase_ball_speed(ball)
                    self.ensure_minimum_ball_speed(ball)

            # Ball hits a brick
            brick_to_remove = -1
            for j, brick in enumerate(self.bricks):
                if ball['rect'].intersects(brick['rect']):
                    # Handle collision
                    self.handleBrickCollision(ball, brick, j)
                    brick_to_remove = j
                    break

            # Remove the brick after bouncing
            if brick_to_remove != -1:
                brick = self.bricks[brick_to_remove]
                
                # Create particles
                self.create_particles(
                    brick['rect'].center().x(),
                    brick['rect'].center().y(),
                    brick['color'],
                    count=4
                )
                
                # Spawn power-up chance
                self.spawn_power_up(brick['rect'].center().x(), brick['rect'].center().y())
                
                # Add score with combo multiplier
                self.combo_count += 1
                score_bonus = int(brick['points'] * (1 + self.combo_count * 0.1))
                self.score += score_bonus
                
                # Remove brick if it's destroyed
                if brick['hits'] <= 1:
                    del self.bricks[brick_to_remove]
                
        # Remove balls that fell off screen
        for i in reversed(balls_to_remove):
            del self.balls[i]
            
        # Check if all balls are gone
        if not self.balls:
            self.lives -= 1
            self.combo_count = 0  # Reset combo
            if self.lives <= 0:
                self.timer.stop()
                self.parent_window.game_over(won=False, score=self.score)
                return
            else:
                # Reset ball for next life
                self.resetBall()
                self.isStarted = False
                
        # Check for power-up collection
        for power_up in self.power_ups[:]:
            if power_up.rect.intersects(self.paddle):
                self.apply_power_up(power_up.power_type)
                self.power_ups.remove(power_up)
                
        # Check laser collisions with bricks
        for laser in self.lasers[:]:
            for j, brick in enumerate(self.bricks):
                if laser.rect.intersects(brick['rect']):
                    # Laser hits brick
                    self.create_particles(
                        brick['rect'].center().x(),
                        brick['rect'].center().y(),
                        brick['color'],
                        count=4
                    )
                    
                    # Add score
                    self.score += int(brick['points'] * 0.5)  # Half points for laser
                    
                    # Damage brick
                    brick['hits'] -= 1
                    if brick['hits'] <= 0:
                        del self.bricks[j]
                        
                    # Remove laser
                    self.lasers.remove(laser)
                    break

        # Check for win condition
        if not self.bricks:
            self.timer.stop()
            self.advance_level()

    def enforce_ball_boundaries(self, ball):
        """Enforce strict boundaries to prevent ball from getting stuck."""
        # Define safety margins
        margin = 2
        
        # Check if ball is outside boundaries and correct position
        if ball['rect'].left() < 0:
            ball['rect'].moveLeft(margin)
            if ball['xDir'] <= 0:
                ball['xDir'] = max(1, abs(ball['xDir']))
                
        elif ball['rect'].right() > self.WIDTH:
            ball['rect'].moveRight(self.WIDTH - margin)
            if ball['xDir'] >= 0:
                ball['xDir'] = min(-1, -abs(ball['xDir']))
                
        if ball['rect'].top() < 0:
            ball['rect'].moveTop(margin)
            if ball['yDir'] <= 0:
                ball['yDir'] = max(1, abs(ball['yDir']))
                
        # Detect if ball is moving too slowly or stopped (potential stuck situation)
        if abs(ball['xDir']) < 1 and abs(ball['yDir']) < 1:
            # Ball is essentially stopped, give it some momentum
            angle = random.uniform(30, 150)  # Random angle between 30-150 degrees
            speed = max(2, int(self.ball_speed_multiplier))
            ball['xDir'] = int(speed * math.cos(math.radians(angle)))
            ball['yDir'] = int(speed * math.sin(math.radians(angle)))
            
        # Detect if ball is in a corner and might be stuck
        corner_margin = 20
        if ((ball['rect'].left() < corner_margin and ball['rect'].top() < corner_margin) or
            (ball['rect'].right() > self.WIDTH - corner_margin and ball['rect'].top() < corner_margin)):
            # Ball is in a top corner, push it toward center
            ball['rect'].moveCenter(QPoint(self.WIDTH // 2, ball['rect'].center().y() + 20))
            ball['xDir'] = random.choice([-1, 1]) * max(1, abs(ball['xDir']))
            ball['yDir'] = max(1, abs(ball['yDir']))

    def ensure_minimum_ball_speed(self, ball):
        """Ensure ball maintains minimum speed to prevent getting stuck."""
        # Check if we have active speed modifiers
        if 'slow_ball' in self.active_power_ups:
            min_speed = max(1, int(self.ball_speed_multiplier * 0.3))  # Lower minimum for slow ball
        elif 'fast_ball' in self.active_power_ups:
            min_speed = max(2, int(self.ball_speed_multiplier * 0.8))  # Higher minimum for fast ball
        else:
            min_speed = max(1, int(self.ball_speed_multiplier * 0.5))
        
        # Ensure minimum horizontal speed
        if abs(ball['xDir']) < min_speed:
            ball['xDir'] = min_speed if ball['xDir'] >= 0 else -min_speed
            
        # Ensure minimum vertical speed  
        if abs(ball['yDir']) < min_speed:
            ball['yDir'] = min_speed if ball['yDir'] >= 0 else -min_speed
            
        # Set maximum speed based on active power-ups
        if 'fast_ball' in self.active_power_ups:
            max_speed = int(6 * self.ball_speed_multiplier)  # Higher max for fast ball
        elif 'slow_ball' in self.active_power_ups:
            max_speed = int(2 * self.ball_speed_multiplier)  # Lower max for slow ball
        else:
            max_speed = int(4 * self.ball_speed_multiplier)
            
        ball['xDir'] = max(-max_speed, min(max_speed, ball['xDir']))
        ball['yDir'] = max(-max_speed, min(max_speed, ball['yDir']))

    def handleBrickCollision(self, ball, brick, brick_index):
        """Handle collision between ball and brick with enhanced anti-stuck measures."""
        # Get the overlap amount to determine bounce direction more accurately
        ball_rect = ball['rect']
        brick_rect = brick['rect']
        
        # Calculate overlap on each side
        overlap_left = ball_rect.right() - brick_rect.left()
        overlap_right = brick_rect.right() - ball_rect.left()
        overlap_top = ball_rect.bottom() - brick_rect.top()
        overlap_bottom = brick_rect.bottom() - ball_rect.top()
        
        # Find the minimum overlap to determine collision side
        min_overlap = min(overlap_left, overlap_right, overlap_top, overlap_bottom)
        
        # Add safety margin for ball repositioning
        safety_margin = 3
        
        # Bounce based on the side with minimum overlap
        if min_overlap == overlap_left and ball['xDir'] > 0:
            ball['xDir'] = -abs(ball['xDir'])  # Ensure negative direction
            ball['rect'].moveRight(brick_rect.left() - safety_margin)
        elif min_overlap == overlap_right and ball['xDir'] < 0:
            ball['xDir'] = abs(ball['xDir'])  # Ensure positive direction
            ball['rect'].moveLeft(brick_rect.right() + safety_margin)
        elif min_overlap == overlap_top and ball['yDir'] > 0:
            ball['yDir'] = -abs(ball['yDir'])  # Ensure negative direction
            ball['rect'].moveBottom(brick_rect.top() - safety_margin)
        elif min_overlap == overlap_bottom and ball['yDir'] < 0:
            ball['yDir'] = abs(ball['yDir'])  # Ensure positive direction
            ball['rect'].moveTop(brick_rect.bottom() + safety_margin)
        else:
            # Fallback: if we can't determine collision side clearly, 
            # reverse both directions and move ball away from brick center
            ball['xDir'] = -ball['xDir']
            ball['yDir'] = -ball['yDir']
            
            # Move ball away from brick center
            brick_center = brick_rect.center()
            ball_center = ball_rect.center()
            
            if ball_center.x() < brick_center.x():
                ball['rect'].moveRight(brick_rect.left() - safety_margin)
            else:
                ball['rect'].moveLeft(brick_rect.right() + safety_margin)
                
            if ball_center.y() < brick_center.y():
                ball['rect'].moveBottom(brick_rect.top() - safety_margin)
            else:
                ball['rect'].moveTop(brick_rect.bottom() + safety_margin)
                
        # Ensure minimum speed after collision
        self.ensure_minimum_ball_speed(ball)
        
        # Increase speed after brick collision for progressive difficulty
        self.increase_ball_speed(ball)
            
        # Damage the brick
        brick['hits'] -= 1
        if brick['hits'] <= 0:
            # Brick is destroyed, will be removed by caller
            pass
        else:
            # Brick still has hits, make it darker
            brick['color'] = brick['color'].darker(120)

    def increase_ball_speed(self, ball):
        """Increase ball speed after collision for progressive difficulty."""
        # Calculate current speed magnitude
        current_speed = math.sqrt(ball['xDir']**2 + ball['yDir']**2)
        
        # Check if we haven't exceeded the maximum speed
        max_allowed_speed = int(4 * self.ball_speed_multiplier * self.max_speed_multiplier)
        
        if current_speed < max_allowed_speed:
            # Increase speed by the factor
            ball['xDir'] = int(ball['xDir'] * self.speed_increase_factor)
            ball['yDir'] = int(ball['yDir'] * self.speed_increase_factor)
            
            # Ensure we don't exceed maximum speed
            new_speed = math.sqrt(ball['xDir']**2 + ball['yDir']**2)
            if new_speed > max_allowed_speed:
                # Scale back to maximum allowed speed
                scale_factor = max_allowed_speed / new_speed
                ball['xDir'] = int(ball['xDir'] * scale_factor)
                ball['yDir'] = int(ball['yDir'] * scale_factor)
                
        # Ensure minimum speed is maintained
        self.ensure_minimum_ball_speed(ball)

    def advance_level(self):
        """Advance to the next level."""
        self.current_level += 1
        
        # Bonus points for completing level
        level_bonus = self.current_level * 500 * self.score_multiplier
        self.score += int(level_bonus)
        
        # Show level complete message
        QMessageBox.information(
            self,
            "Level Complete!",
            f"Level {self.current_level - 1} Complete!\n"
            f"Level Bonus: {int(level_bonus)} points\n"
            f"Current Score: {self.score}\n\n"
            f"Starting Level {self.current_level}..."
        )
        
        # Reset for next level
        self.initBricks()
        self.resetBall()
        self.isStarted = False
        self.combo_count = 0
        
        # Clear active power-ups for new level
        self.active_power_ups.clear()
        self.power_up_timers.clear()
        self.power_ups.clear()
        self.lasers.clear()
        
        # Reset paddle to original size
        self.paddle = QRect(self.paddle.x(), self.paddle.y(), 
                            self.original_paddle_width, self.PADDLE_HEIGHT)


class BreakoutGame(QMainWindow):
    """
    Main application window for the Breakout game.
    """
    def __init__(self, main_window, parent=None):
        super().__init__()
        self.main_window = main_window
        self.high_score_manager = HighScoreManager()
        
        # Remove minimize and maximize buttons
        self.setWindowFlags(Qt.Window |
                            Qt.CustomizeWindowHint |
                            Qt.WindowTitleHint |
                            Qt.WindowCloseButtonHint)
        
        # Set the window icon and title
        self.setWindowIcon(get_icon("breakout.png"))
        self.title = "Breakout Game"
        
        # Set the window opacity
        self.opacity = 1
        
        # Initialize the game board
        self.board = None
        self.difficulty = "Medium"  # Default difficulty
        # For title animation
        self.title_timer = 0
        

    def init_ui(self):
        """Set up the user interface dimensions and appearance."""
        self.setFixedSize(self.board.WIDTH, self.board.HEIGHT)
        self.setWindowTitle(self.title)
        self.setWindowOpacity(self.opacity)
        self.show()

    def init_game(self):
        """Initialize the game board with selected difficulty."""
        self.board = Board(self, self.difficulty)
        self.setCentralWidget(self.board)
        self.title_timer = 10  # Show title for 0.5 seconds

    def start_game(self):
        """Start the game by initializing the game window and UI."""
        # Stop and disconnect any existing timers before starting a new game
        if hasattr(self, 'board') and self.board:
            if hasattr(self.board, 'timer') and self.board.timer.isActive():
                self.board.timer.stop()
            
            # Disconnect all previous timer connections
            try:
                if hasattr(self.board, 'timer'):
                    self.board.timer.timeout.disconnect()
            except TypeError:
                # No connections to disconnect
                pass
                
            if hasattr(self.board, 'effect_timer') and self.board.effect_timer.isActive():
                self.board.effect_timer.stop()
            
            try:
                if hasattr(self.board, 'effect_timer'):
                    self.board.effect_timer.timeout.disconnect()
            except TypeError:
                # No connections to disconnect
                pass
        
        self.difficulty = "Hard"
        
        # Show welcome dialog with instructions
        welcome_msg = (
            f"Welcome to Breakout - {self.difficulty} Mode!\n\n"
            "Features:\n"
            " - Multiple difficulty levels with score multipliers\n"
            " - Power-ups: Multiball, Paddle size, Sticky paddle, Lasers, etc.\n"
            " - Particle effects and visual enhancements\n"
            " - High score tracking\n"
            " - Progressive levels with increasing difficulty\n"
            " - Lives system and combo scoring\n\n"
            "Controls:\n"
            " - Use 'A' to move paddle left\n"
            " - Use 'D' to move paddle right\n"
            " - Press SPACE to start game or shoot lasers\n"
            " - Press P to pause/unpause\n\n"
            "Click 'OK' to start playing!"
        )
        if QMessageBox.information(self, 
                                   "Welcome to Breakout", 
                                   welcome_msg, QMessageBox.Ok) == QMessageBox.Ok:
            self.init_game()                   # Create board and set dimensions.
            self.init_ui()                     # Set up UI dimensions.
        else:
            self.close()

    def game_over(self, won=False, score=0):
        """Display game over message and handle restart/close."""
        # Stop the game timer immediately to prevent multiple dialogs
        if hasattr(self, 'board') and self.board and \
           hasattr(self.board, 'timer') and self.board.timer.isActive():
            self.board.timer.stop()
        
        # Stop effect timer as well
        if hasattr(self, 'board') and self.board and \
           hasattr(self.board, 'effect_timer') and self.board.effect_timer.isActive():
            self.board.effect_timer.stop()
        
        # Check if it's a high score
        is_high_score = self.high_score_manager.add_score(self.difficulty, score)
        
        # Get level before board might be set to None
        level_reached = self.board.current_level if self.board else 0
        
        if won:
            title = "Congratulations!"
            message = f"You completed all levels on {self.difficulty} mode!\n\n"
        else:
            title = "Game Over"
            message = f"Game over on {self.difficulty} mode!\n\n"
            
        message += f"Final Score: {score}\n"
        message += f"Level Reached: {level_reached}\n"
        
        if is_high_score:
            message += "\n🎉 NEW HIGH SCORE! 🎉\n"
            
        message += "\nDo you want to play again?"
        
        # Store the dialog response in a variable
        reply = QMessageBox.question(
            self,
            title,
            message,
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No  # Default to "No" to prevent accidental restarts
        )
        
        if reply == QMessageBox.Yes:
            self.start_game()
        else:
            # Clean up completely and close
            self.cleanup_and_close()
            
    def end_game(self):
        """End the game by stopping the timer and closing the window."""
        self.cleanup_and_close()

    def cleanup_and_close(self):
        """Completely clean up and close the game window."""
        # Stop all board timers and disconnect signals
        if hasattr(self, 'board') and self.board:
            if hasattr(self.board, 'timer') and self.board.timer.isActive():
                self.board.timer.stop()
            
            # Disconnect all timer connections to prevent memory leaks
            try:
                if hasattr(self.board, 'timer'):
                    self.board.timer.timeout.disconnect()
            except TypeError:
                # No connections to disconnect
                pass
                
            if hasattr(self.board, 'effect_timer') and self.board.effect_timer.isActive():
                self.board.effect_timer.stop()
            
            try:
                if hasattr(self.board, 'effect_timer'):
                    self.board.effect_timer.timeout.disconnect()
            except TypeError:
                # No connections to disconnect
                pass
        
        # Clear references
        self.board = None
        
        # Close the window
        self.close()
        
        # Clear any reference from main window
        if hasattr(self, 'main_window') and self.main_window:
            # If the main window has a reference to this game, clear it
            if hasattr(self.main_window, 'current_game'):
                self.main_window.current_game = None
                
    def closeEvent(self, event):
        """Handle the window close event."""
        # Ensure proper cleanup when window is closed
        self.cleanup_and_close()
        event.accept()


import warnings

from PyQt5.QtCore import QTimer, QObject

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class AnimationManager(QObject):
    """
    Manages a single, global timer to handle animations for all
    registered objects (like Annotations and WorkAreas).
    
    This avoids creating thousands of QTimer objects, which can
    exhaust system resources and crash the application.
    """
    
    def __init__(self, parent=None):
        """Initialize the manager."""
        super().__init__(parent)
        
        # The set of all objects that are currently animating
        self.animating_objects = set()
        
        # The single, global timer that drives all animations
        self.global_animation_timer = QTimer(self)
        self.global_animation_timer.timeout.connect(self._update_all_animations)

    def start_timer(self, interval=50):
        """
        Starts the global animation timer.
        
        Args:
            interval (int): The interval in milliseconds (e.g., 50ms = 20 FPS).
        """
        if not self.global_animation_timer.isActive():
            self.global_animation_timer.start(interval)
            
    def stop_timer(self):
        """Stops the global animation timer."""
        self.global_animation_timer.stop()

    def register_animating_object(self, obj):
        """
        Add an object to the set of items to be animated.
        
        Args:
            obj: The object to animate. Must have:
                 - is_graphics_item_valid() -> bool
                 - tick_animation()
        """
        self.animating_objects.add(obj)
        
    def unregister_animating_object(self, obj):
        """
        Remove an object from the set of items to be animated.
        
        Args:
            obj: The object to stop animating.
        """
        self.animating_objects.discard(obj)

    def _update_all_animations(self):
        """
        This is the engine, called by the timer on every tick.
        It loops through all registered objects and calls their
        tick_animation() method.
        """
        if not self.animating_objects:
            return
            
        # Loop over a static copy (list) of the set.
        # This allows objects to be safely unregistered during iteration
        # (e.g., if they are deleted).
        for obj in list(self.animating_objects):
            try:
                # 1. Safely check if the object is still "alive"
                #    by checking its graphics item.
                if obj and obj.is_graphics_item_valid():
                    # 2. If alive, call its public animation method
                    obj.tick_animation()
                else:
                    # 3. If it's not valid (e.g., deleted), unregister it
                    self.unregister_animating_object(obj)
            
            except (RuntimeError, AttributeError):
                # 4. Catch any other errors (e.g., C++ object deleted)
                #    and unregister the object to prevent further errors.
                self.unregister_animating_object(obj)


import warnings

import os

import numpy as np

from PyQt5.QtCore import Qt, QRectF
from PyQt5.QtGui import QPen, QColor, QPainter
from PyQt5.QtWidgets import QGraphicsObject, QStyle, QVBoxLayout, QLabel, QWidget, QGraphicsItem

warnings.filterwarnings("ignore", category=DeprecationWarning)


# ----------------------------------------------------------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------------------------------------------------------

POINT_SIZE = 15
POINT_WIDTH = 3
SPRITE_SIZE = 32
ANNOTATION_WIDTH = 5


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class EmbeddingPointItem(QGraphicsObject):
    """
    A custom QGraphicsObject that can display as a dot or an image sprite,
    getting its state from an associated AnnotationDataItem.
    """

    def __init__(self, data_item, viewer):
        """
        Initializes the point item.
        Args:
            data_item (AnnotationDataItem): The data item that holds the state.
            viewer (EmbeddingViewer): A reference to the parent viewer.
        """
        super(EmbeddingPointItem, self).__init__()

        self.data_item = data_item
        self.viewer = viewer
        self.thumbnail_pixmap = None

        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemIgnoresTransformations, True)

        self.default_pen = QPen(QColor("black"), POINT_WIDTH)
        self.setPos(self.data_item.embedding_x, self.data_item.embedding_y)
        self.setToolTip(self.data_item.get_tooltip_text())
        
        # --- Animation Properties ---
        self.animation_manager = None
        self.is_animating = False
        
        # --- Animation properties ---
        self._pulse_alpha = 128  # Starting alpha for pulsing (semi-transparent)
        self._pulse_direction = 1  # 1 for increasing alpha, -1 for decreasing
        
    def set_animation_manager(self, manager):
        """
        Binds this object to the central AnimationManager.
        
        Args:
            manager (AnimationManager): The central animation manager instance.
        """
        self.animation_manager = manager
        
    def is_graphics_item_valid(self):
        """
        Checks if the graphics item is still valid and added to a scene.
        
        Returns:
            bool: True if the item exists and has a scene, False otherwise.
        """
        try:
            return self.scene() is not None
        except RuntimeError:
            # This can happen if the C++ part of the item is deleted
            return False

    def boundingRect(self):
        """Returns the bounding rectangle, which depends on the display mode and depth."""
        
        scale_factor = 1.0
        if self.viewer and self.viewer.is_3d_data and self.viewer.z_range > 0:
            # Normalize z from its global range to a [0, 1] range
            z_normalized = (self.data_item.embedding_z - self.viewer.min_z) / self.viewer.z_range
            # Map normalized z to a scale factor (e.g., from 0.5x to 1.5x)
            scale_factor = 0.5 + z_normalized
    
        if self.viewer and self.viewer.display_mode == 'sprites':
            ar = self.data_item.aspect_ratio
            if ar >= 1.0:
                width = SPRITE_SIZE * scale_factor
                height = (SPRITE_SIZE / ar) * scale_factor
            else:
                height = SPRITE_SIZE * scale_factor
                width = (SPRITE_SIZE * ar) * scale_factor
            return QRectF(0, 0, width, height)
        else:
            
            size = POINT_SIZE * scale_factor
            return QRectF(0, 0, size, size)

    def update_tooltip(self):
        """Updates the tooltip by fetching the latest text from the data item."""
        self.setToolTip(self.data_item.get_tooltip_text())

    def paint(self, painter, option, widget):
        """
        Custom paint method to draw either a dot or a sprite with a border.
        """
        option.state &= ~QStyle.State_Selected
        painter.setRenderHint(QPainter.Antialiasing)
        
        # Calculate scale_factor for size and opacity
        scale_factor = 1.0
        if self.viewer and self.viewer.is_3d_data and self.viewer.z_range > 0:
            z_normalized = (self.data_item.embedding_z - self.viewer.min_z) / self.viewer.z_range
            scale_factor = 0.5 + z_normalized  # From 0.5x to 1.5x

        # Calculate scaled pen width for borders (clamp to avoid extremes)
        scaled_pen_width = max(1, min(POINT_WIDTH * scale_factor, 6))  # Clamp between 1 and 6 for usability
        
        # Calculate opacity
        opacity = 255
        if self.viewer and self.viewer.is_3d_data and self.viewer.z_range > 0:
            z_normalized = (self.data_item.embedding_z - self.viewer.min_z) / self.viewer.z_range
            opacity = int(128 + 127 * z_normalized)

        base_color = self.data_item.effective_color
        effective_brush_color = QColor(base_color)
        effective_brush_color.setAlpha(opacity)
        
        display_mode = self.viewer.display_mode if self.viewer else 'dots'

        if display_mode == 'sprites':
            # Ensure the pixmap is scaled to the current boundingRect size (handles dynamic scaling during rotation)
            current_size = self.boundingRect().size().toSize()
            if self.thumbnail_pixmap is None or self.thumbnail_pixmap.size() != current_size:
                source_pixmap = self.data_item.annotation.get_cropped_image_graphic()
                if source_pixmap and not source_pixmap.isNull():
                    self.thumbnail_pixmap = source_pixmap.scaled(
                        current_size,
                        Qt.KeepAspectRatio, Qt.SmoothTransformation
                    )
            
            if self.thumbnail_pixmap:
                painter.drawPixmap(self.boundingRect().topLeft(), self.thumbnail_pixmap)

            # Scaled border pen for sprites
            border_color = QColor(self.data_item.effective_color)
            border_color.setAlpha(opacity)
            border_pen = QPen(border_color, scaled_pen_width)
            if self.isSelected():
                # Tell manager to start animating
                self.animate()
                
                border_color = QColor(self.data_item.effective_color).darker(150)  
                border_color.setAlpha(self._pulse_alpha)
                border_pen = QPen(border_color, scaled_pen_width)
                border_pen.setStyle(Qt.DotLine)
            else:
                # Tell manager to stop animating
                self.deanimate()

            painter.setPen(border_pen)
            painter.setBrush(Qt.NoBrush)
            painter.drawRect(self.boundingRect())
        else:
            # Draw Original Dot with scaled size and pen
            if self.isSelected():
                # Tell manager to start animating
                self.animate()
                
                darker_color = QColor(self.data_item.effective_color).darker(150)  
                darker_color.setAlpha(self._pulse_alpha)
                animated_pen = QPen(darker_color, scaled_pen_width)
                animated_pen.setStyle(Qt.DotLine)
                painter.setPen(animated_pen)
            else:
                # Tell manager to stop animating
                self.deanimate()
                
                pen_color = QColor("black")
                pen_color.setAlpha(opacity)
                painter.setPen(QPen(pen_color, scaled_pen_width))

            painter.setBrush(effective_brush_color)
            painter.drawEllipse(self.boundingRect())
            
    def tick_animation(self):
        """
        Perform one 'tick' of the animation.
        This is the public entry point for the global manager.
        """
        # This just calls the existing private method that holds the logic
        self._update_pulse_alpha()
    
    def _update_pulse_alpha(self):
        """Update the pulse alpha for a heartbeat-like effect: quick rise, slow fall."""
        if self._pulse_direction == 1:
            # Quick increase (systole-like)
            self._pulse_alpha += 30
        else:
            # Slow decrease (diastole-like)
            self._pulse_alpha -= 10  # <-- Corrected from += to -=

        # Check direction before clamping to ensure smooth transition
        if self._pulse_alpha >= 255:
            self._pulse_alpha = 255  # Clamp to max
            self._pulse_direction = -1
        elif self._pulse_alpha <= 50:
            self._pulse_alpha = 50   # Clamp to min
            self._pulse_direction = 1
        
        self.update()  # Trigger repaint
    
    def animate(self):
        """Start the pulsing animation by registering with the global timer."""
        self.is_animating = True
        if self.animation_manager:
            self.animation_manager.register_animating_object(self)
            
    def deanimate(self):
        """Stop the pulsing animation by de-registering from the global timer."""
        self.is_animating = False
        if self.animation_manager:
            self.animation_manager.unregister_animating_object(self)
            
        self._pulse_alpha = 128  # Reset to default
        self.update()  # Apply the default style
    
    def __del__(self):
        """Clean up the timer when the item is deleted."""
        if hasattr(self, 'is_animating') and self.is_animating:
            self.deanimate()
            

class AnnotationImageWidget(QWidget):
    """Widget to display a single annotation image crop with selection support."""

    def __init__(self, data_item, widget_height=96, annotation_viewer=None, parent=None):
        super(AnnotationImageWidget, self).__init__(parent)
        self.data_item = data_item
        self.annotation = data_item.annotation
        self.annotation_viewer = annotation_viewer

        self.widget_height = widget_height
        self.aspect_ratio = 1.0
        self.pixmap = None
        self.is_loaded = False

        # --- Animation Properties ---
        self.animation_manager = None
        self.is_animating = False
        
        # --- Animation properties (no timer) ---
        self._pulse_alpha = 128
        self._pulse_direction = 1

        self.setup_ui()
        self.recalculate_aspect_ratio()
        self.update_height(self.widget_height)
        self.update_tooltip()
        
    def set_animation_manager(self, manager):
        """
        Binds this object to the central AnimationManager.
        
        Args:
            manager (AnimationManager): The central animation manager instance.
        """
        self.animation_manager = manager
        
    def is_graphics_item_valid(self):
        """
        Checks if the widget is still valid and visible.
        
        Returns:
            bool: True if the widget is visible, False otherwise.
        """
        try:
            return self.isVisible()
        except RuntimeError:
            return False

    def setup_ui(self):
        """Set up the basic UI with a label for the image."""
        layout = QVBoxLayout(self)
        layout.setContentsMargins(4, 4, 4, 4)
        self.image_label = QLabel()
        self.image_label.setAlignment(Qt.AlignCenter)
        self.image_label.setScaledContents(True)
        self.image_label.setStyleSheet("border: none;")
        layout.addWidget(self.image_label)
        
    def update_tooltip(self):
        """Updates the tooltip by fetching the latest text from the data item."""
        self.setToolTip(self.data_item.get_tooltip_text())

    def recalculate_aspect_ratio(self):
        """Calculate and store the annotation's aspect ratio."""
        annotation = self.data_item.annotation
        
        # Try to use the cropped_bbox attribute first
        if hasattr(annotation, 'cropped_bbox'):
            min_x, min_y, max_x, max_y = annotation.cropped_bbox
            width = max_x - min_x
            height = max_y - min_y
            
            if height > 0:
                self.aspect_ratio = width / height
                return
        
        # Fallback to bounding box methods
        try:
            top_left = annotation.get_bounding_box_top_left()
            bottom_right = annotation.get_bounding_box_bottom_right()
            
            if top_left and bottom_right:
                width = bottom_right.x() - top_left.x()
                height = bottom_right.y() - top_left.y()
                
                if height > 0:
                    self.aspect_ratio = width / height
                    return
        except (AttributeError, TypeError):
            pass
        
        # Last resort: try to get aspect ratio from the cropped image
        try:
            pixmap = annotation.get_cropped_image()
            if pixmap and not pixmap.isNull() and pixmap.height() > 0:
                self.aspect_ratio = pixmap.width() / pixmap.height()
                return
        except (AttributeError, TypeError):
            pass
        
        # Default to square if we can't determine aspect ratio
        self.aspect_ratio = 1.0

    def load_image(self):
        """Loads the image pixmap if it hasn't been loaded yet."""
        if self.is_loaded:
            return

        try:
            cropped_pixmap = self.annotation.get_cropped_image_graphic()
            if cropped_pixmap and not cropped_pixmap.isNull():
                self.pixmap = cropped_pixmap
                self.is_loaded = True
                self._display_pixmap()
            else:
                self.image_label.setText("No Image\nAvailable")
                self.pixmap = None
        except Exception as e:
            print(f"Error loading annotation image: {e}")
            self.image_label.setText("Error\nLoading Image")
            self.pixmap = None

    def unload_image(self):
        """Unloads the pixmap to free memory."""
        if not self.is_loaded:
            return
        self.pixmap = None
        self.image_label.clear()
        self.is_loaded = False

    def _display_pixmap(self):
        """Scales and displays the currently loaded pixmap."""
        if self.pixmap:
            new_width = int(self.widget_height * self.aspect_ratio)
            scaled_pixmap = self.pixmap.scaled(
                new_width - 8,
                self.widget_height - 8,
                Qt.KeepAspectRatio,
                Qt.SmoothTransformation
            )
            self.image_label.setPixmap(scaled_pixmap)

    def update_height(self, new_height):
        """Updates the widget's height and rescales its width and content accordingly."""
        self.widget_height = new_height
        new_width = int(self.widget_height * self.aspect_ratio)
        self.setFixedSize(new_width, new_height)
        if self.pixmap:
            self._display_pixmap()
        self.update()

    def update_selection_visuals(self):
        """
        Updates the widget's visual state based on the data_item's selection
        status. This should be called by the controlling viewer.
        """
        is_selected = self.data_item.is_selected

        if is_selected:
            self.animate()
        else:
            self.deanimate()
            self._pulse_alpha = 128  # Reset to default

        # Trigger a repaint to show the new selection state (border, etc.)
        self.update()

    def is_selected(self):
        """Return whether this widget is selected via the data item."""
        return self.data_item.is_selected

    def _update_animation_frame(self):
        """Update the animation offset and schedule a repaint."""
        # Removed: self.animation_offset = (self.animation_offset + 1) % 20
        self.update()

    def paintEvent(self, event):
        """Handle custom drawing for the widget, including the selection border."""
        super().paintEvent(event)
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        effective_label = self.data_item.effective_label
        pen_color = self.data_item.effective_color
        if effective_label and effective_label.id == "-1":
            # If the label is a temporary one (e.g., "-1", Review), use black for the pen color
            pen_color = QColor("black")

        if self.is_selected():
            # Use a darker version of the color for better visibility
            pen_color = pen_color.darker(150)  # Changed to darker for brighter selected appearance
            pen_color.setAlpha(self._pulse_alpha)  # Apply pulsing alpha for animation
            pen = QPen(pen_color, ANNOTATION_WIDTH)
            pen.setStyle(Qt.DotLine)  # Predefined dotted line (static, no movement)
        else:
            pen = QPen(pen_color, ANNOTATION_WIDTH)
            pen.setStyle(Qt.SolidLine)

        painter.setPen(pen)
        painter.setBrush(Qt.NoBrush)

        width = pen.width()
        half_width = (width - 1) // 2
        rect = self.rect().adjusted(half_width, half_width, -half_width, -half_width)
        painter.drawRect(rect)
        
    def tick_animation(self):
        """
        Perform one 'tick' of the animation.
        This is the public entry point for the global manager.
        """
        # This just calls the existing private method that holds the logic
        self._update_pulse_alpha()
        
    def _update_pulse_alpha(self):
        """Update the pulse alpha for a heartbeat-like effect: quick rise, slow fall."""
        if self._pulse_direction == 1:
            # Quick increase (systole-like)
            self._pulse_alpha += 30
        else:
            # Slow decrease (diastole-like)
            self._pulse_alpha -= 10  # <-- Corrected from += to -=

        # Check direction before clamping to ensure smooth transition
        if self._pulse_alpha >= 255:
            self._pulse_alpha = 255  # Clamp to max
            self._pulse_direction = -1
        elif self._pulse_alpha <= 50:
            self._pulse_alpha = 50   # Clamp to min
            self._pulse_direction = 1
        
        self.update()  # Trigger repaint
        
    def animate(self):
        """Start the pulsing animation by registering with the global timer."""
        self.is_animating = True
        if self.animation_manager:
            self.animation_manager.register_animating_object(self)
            
    def deanimate(self):
        """Stop the pulsing animation by de-registering from the global timer."""
        self.is_animating = False
        if self.animation_manager:
            self.animation_manager.unregister_animating_object(self)
            
        self._pulse_alpha = 128  # Reset to default
        self.update()  # Apply the default style
    
    def mousePressEvent(self, event):
        """Handle mouse press events for selection, delegating logic to the viewer."""
        if event.button() == Qt.LeftButton:
            if self.annotation_viewer and hasattr(self.annotation_viewer, 'handle_annotation_selection'):
                # The viewer is the controller and will decide how to change the selection state
                self.annotation_viewer.handle_annotation_selection(self, event)
        elif event.button() == Qt.RightButton:
            if self.annotation_viewer and hasattr(self.annotation_viewer, 'handle_annotation_context_menu'):
                self.annotation_viewer.handle_annotation_context_menu(self, event)
                event.accept()
                return
            else:
                event.ignore()
        super().mousePressEvent(event)
        
    def __del__(self):
        """Clean up the timer when the widget is deleted."""
        if hasattr(self, 'is_animating') and self.is_animating:
            self.deanimate()            


class AnnotationDataItem:
    """
    Holds all annotation state information for consistent display across viewers.
    This acts as the "ViewModel" for a single annotation, serving as the single
    source of truth for its state in the UI.
    """

    def __init__(self, annotation, embedding_x=None, embedding_y=None, embedding_id=None):
        self.annotation = annotation
        
        self.embedding_x = embedding_x if embedding_x is not None else 0.0
        self.embedding_y = embedding_y if embedding_y is not None else 0.0
        
        self.embedding_z = 0.0  # This will store the rotated Z-value (depth)
        
        # Store the original, un-rotated 3D coordinates from the embedding
        self.embedding_x_3d = 0.0
        self.embedding_y_3d = 0.0
        self.embedding_z_3d = 0.0
        
        self.embedding_id = embedding_id
        
        self._is_selected = False
        self._preview_label = None
        self._original_label = annotation.label

        # Calculate and store aspect ratio on initialization
        self.aspect_ratio = self._calculate_aspect_ratio()
        
        # To store pre-formatted top-k prediction details
        self.prediction_details = None
        # To store prediction probabilities for sorting
        self.prediction_probabilities = None

    def _calculate_aspect_ratio(self):
        """Calculate and return the annotation's aspect ratio."""
        annotation = self.annotation
        
        if hasattr(annotation, 'cropped_bbox'):
            min_x, min_y, max_x, max_y = annotation.cropped_bbox
            width = max_x - min_x
            height = max_y - min_y
            if height > 0:
                return width / height

        try:
            top_left = annotation.get_bounding_box_top_left()
            bottom_right = annotation.get_bounding_box_bottom_right()
            if top_left and bottom_right:
                width = bottom_right.x() - top_left.x()
                height = bottom_right.y() - top_left.y()
                if height > 0:
                    return width / height
        except (AttributeError, TypeError):
            pass

        try:
            pixmap = annotation.get_cropped_image()
            if pixmap and not pixmap.isNull() and pixmap.height() > 0:
                return pixmap.width() / pixmap.height()
        except (AttributeError, TypeError):
            pass
        
        return 1.0  # Default to square
        
    @property
    def effective_label(self):
        """Get the current effective label (preview if it exists, otherwise original)."""
        return self._preview_label if self._preview_label else self.annotation.label

    @property
    def effective_color(self):
        """Get the effective color for this annotation based on the effective label."""
        return self.effective_label.color

    @property
    def is_selected(self):
        """Check if this annotation is selected."""
        return self._is_selected

    def set_selected(self, selected):
        """Set the selection state. This is the single point of control."""
        self._is_selected = selected

    def set_preview_label(self, label):
        """Set a preview label for this annotation."""
        self._preview_label = label

    def clear_preview_label(self):
        """Clear the preview label and revert to the original."""
        self._preview_label = None

    def has_preview_changes(self):
        """Check if this annotation has a temporary preview label assigned."""
        return self._preview_label is not None

    def apply_preview_permanently(self):
        """Apply the preview label permanently to the underlying annotation object."""
        if self._preview_label:
            self.annotation.update_label(self._preview_label)
            self.annotation.update_user_confidence(self._preview_label)
            self._original_label = self._preview_label
            self._preview_label = None
            return True
        return False

    def get_display_info(self):
        """Get display information for this annotation."""
        return {
            'id': self.annotation.id,
            'label': self.effective_label.short_label_code,
            'confidence': self.get_effective_confidence(),
            'type': type(self.annotation).__name__,
            'image': os.path.basename(self.annotation.image_path),
            'embedding_id': self.embedding_id,
            'color': self.effective_color
        }
    
    def get_tooltip_text(self):
        """
        Generates a rich HTML-formatted tooltip with all relevant information.
        """
        info = self.get_display_info()
        
        tooltip_parts = [
            f"<b>ID:</b> {info['id']}",
            f"<b>Image:</b> {info['image']}",
            f"<b>Label:</b> {info['label']}",
            f"<b>Type:</b> {info['type']}"
        ]

        # Add prediction details if they exist
        if self.prediction_details:
            tooltip_parts.append(f"<hr>{self.prediction_details}")

        return "<br>".join(tooltip_parts)

    def get_effective_confidence(self):
        """
        Get the effective confidence value, handling scalar, array, and vector predictions.
        """
        # First check if prediction probabilities are available from model predictions
        if hasattr(self, 'prediction_probabilities') and self.prediction_probabilities is not None:
            probs = self.prediction_probabilities
            try:
                # This will succeed for lists and multi-element numpy arrays
                if len(probs) > 0:
                    return float(np.max(probs))
            except TypeError:
                # This will catch the error if `len()` is called on a scalar or 0-D array.
                # In this case, the value of `probs` itself is the confidence score.
                return float(probs)

        # Fallback to existing confidence values
        if self.annotation.verified and hasattr(self.annotation, 'user_confidence') and self.annotation.user_confidence:
            return list(self.annotation.user_confidence.values())[0]
        elif hasattr(self.annotation, 'machine_confidence') and self.annotation.machine_confidence:
            return list(self.annotation.machine_confidence.values())[0]
            
        return 0.0


import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

import os 

from PyQt5.QtWidgets import (QVBoxLayout, QLabel, QGroupBox, QFormLayout,
                             QDoubleSpinBox, QComboBox, QSpinBox, QHBoxLayout,
                             QWidget, QStackedWidget, QGridLayout, QMessageBox,
                             QDialog, QListWidget, QPushButton, QFileDialog,
                             QGraphicsView)
from PyQt5.QtCore import Qt                           

from coralnet_toolbox.Icons import get_icon


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class TileSizeInput(QGroupBox):
    def __init__(self, parent=None):
        super().__init__("Tile Size", parent)
        layout = QFormLayout(self)

        # Width input in pixels
        self.width_spin = QSpinBox()
        self.width_spin.setRange(32, 10000)
        self.width_spin.setValue(1024)

        # Height input in pixels
        self.height_spin = QSpinBox()
        self.height_spin.setRange(32, 10000)
        self.height_spin.setValue(768)

        layout.addRow("Width:", self.width_spin)
        layout.addRow("Height:", self.height_spin)
        
    def get_value(self):
        """
        Get the current tile size values.
        
        Returns:
            tuple: (width, height)
        """
        return self.width_spin.value(), self.height_spin.value()

    def get_sizes(self, image_width=None, image_height=None, validate=True):
        """
        Get tile sizes with optional validation against image dimensions.
        
        Args:
            image_width (int, optional): Width of source image
            image_height (int, optional): Height of source image
            validate (bool): Whether to validate against image dimensions
            
        Returns:
            tuple: (width, height) if valid
            None: if invalid when validation enabled
        """
        tile_size = self.get_value()
        
        try:
            # Basic type validation always performed
            if not isinstance(tile_size, tuple) or len(tile_size) != 2:
                raise ValueError("Tile size must be a tuple of two values")
                
            width, height = tile_size
            
            # Size validation only if requested and dimensions provided
            if validate and image_width and image_height:
                if width > image_width:
                    raise ValueError(f"Tile width ({width}) cannot exceed image width ({image_width})")
                if height > image_height:
                    raise ValueError(f"Tile height ({height}) cannot exceed image height ({image_height})")
        
            return width, height
            
        except ValueError as e:
            if validate:
                QMessageBox.warning(self, "Invalid Tile Size", str(e))
                return None
            return tile_size